import pytest

from src.core.config import BatchAgentConfig, BatchCheckpointerConfig, BatchLLMConfig


@pytest.mark.asyncio
async def test_llm_config_old_format(temp_dir):
    config_file = temp_dir / "config.llms.yml"
    config_file.write_text(
        """
llms:
  - model: test-model
    alias: test-llm
    provider: anthropic
    max_tokens: 1000
    temperature: 0.1
"""
    )

    config = await BatchLLMConfig.from_yaml(file_path=config_file)
    assert len(config.llms) == 1
    assert config.llms[0].alias == "test-llm"


@pytest.mark.asyncio
async def test_llm_config_new_format(temp_dir):
    llms_dir = temp_dir / "llms"
    llms_dir.mkdir()

    # Files can be organized by provider
    (llms_dir / "anthropic.yml").write_text(
        """
- model: test-model
  alias: test-llm
  provider: anthropic
  max_tokens: 1000
  temperature: 0.1
"""
    )

    config = await BatchLLMConfig.from_yaml(dir_path=llms_dir)
    assert len(config.llms) == 1
    assert config.llms[0].alias == "test-llm"


@pytest.mark.asyncio
async def test_llm_config_combine_both_formats(temp_dir):
    """Test that unique LLMs from both file and directory can be loaded together."""
    config_file = temp_dir / "config.llms.yml"
    config_file.write_text(
        """
llms:
  - model: old-model
    alias: old-llm
    provider: anthropic
    max_tokens: 1000
    temperature: 0.1
  - model: file-model
    alias: file-llm
    provider: openai
    max_tokens: 2000
    temperature: 0.2
"""
    )

    llms_dir = temp_dir / "llms"
    llms_dir.mkdir()
    (llms_dir / "anthropic.yml").write_text(
        """
- model: new-model
  alias: new-llm
  provider: anthropic
  max_tokens: 1000
  temperature: 0.1
- model: dir-model
  alias: dir-llm
  provider: anthropic
  max_tokens: 3000
  temperature: 0.3
"""
    )

    config = await BatchLLMConfig.from_yaml(file_path=config_file, dir_path=llms_dir)
    assert len(config.llms) == 4

    llm_dict = {llm.alias: llm for llm in config.llms}
    assert "old-llm" in llm_dict
    assert "file-llm" in llm_dict
    assert "new-llm" in llm_dict
    assert "dir-llm" in llm_dict


@pytest.mark.asyncio
async def test_llm_config_duplicate_between_formats_raises_error(temp_dir):
    """Test that duplicate LLMs between file and directory raise an error."""
    config_file = temp_dir / "config.llms.yml"
    config_file.write_text(
        """
llms:
  - model: old-model
    alias: duplicate-llm
    provider: anthropic
    max_tokens: 1000
    temperature: 0.1
"""
    )

    llms_dir = temp_dir / "llms"
    llms_dir.mkdir()
    (llms_dir / "openai.yml").write_text(
        """
- model: new-model
  alias: duplicate-llm
  provider: openai
  max_tokens: 2000
  temperature: 0.2
"""
    )

    with pytest.raises(ValueError, match=r"Duplicate llm 'alias': 'duplicate-llm'"):
        await BatchLLMConfig.from_yaml(file_path=config_file, dir_path=llms_dir)


@pytest.mark.asyncio
async def test_checkpointer_config_old_format(temp_dir):
    config_file = temp_dir / "config.checkpointers.yml"
    config_file.write_text(
        """
checkpointers:
  - type: sqlite
    max_connections: 10
"""
    )

    config = await BatchCheckpointerConfig.from_yaml(file_path=config_file)
    assert len(config.checkpointers) == 1
    assert config.checkpointers[0].type == "sqlite"


@pytest.mark.asyncio
async def test_checkpointer_config_new_format(temp_dir):
    checkpointers_dir = temp_dir / "checkpointers"
    checkpointers_dir.mkdir()

    (checkpointers_dir / "sqlite.yml").write_text(
        """
type: sqlite
max_connections: 10
"""
    )

    config = await BatchCheckpointerConfig.from_yaml(dir_path=checkpointers_dir)
    assert len(config.checkpointers) == 1
    assert config.checkpointers[0].type == "sqlite"


@pytest.mark.asyncio
async def test_agent_config_old_format(temp_dir, mock_llm_config):
    config_file = temp_dir / "config.agents.yml"
    config_file.write_text(
        f"""
agents:
  - name: test-agent
    prompt: "test prompt"
    llm: {mock_llm_config.alias}
    default: true
"""
    )

    llm_config = BatchLLMConfig(llms=[mock_llm_config])
    config = await BatchAgentConfig.from_yaml(
        file_path=config_file, batch_llm_config=llm_config
    )
    assert len(config.agents) == 1
    assert config.agents[0].name == "test-agent"


@pytest.mark.asyncio
async def test_agent_config_new_format(temp_dir, mock_llm_config):
    agents_dir = temp_dir / "agents"
    agents_dir.mkdir()

    (agents_dir / "test-agent.yml").write_text(
        f"""
name: test-agent
prompt: "test prompt"
llm: {mock_llm_config.alias}
default: true
"""
    )

    llm_config = BatchLLMConfig(llms=[mock_llm_config])
    config = await BatchAgentConfig.from_yaml(
        dir_path=agents_dir, batch_llm_config=llm_config
    )
    assert len(config.agents) == 1
    assert config.agents[0].name == "test-agent"


@pytest.mark.asyncio
async def test_agent_config_merge_both_formats(temp_dir, mock_llm_config):
    config_file = temp_dir / "config.agents.yml"
    config_file.write_text(
        f"""
agents:
  - name: old-agent
    prompt: "old prompt"
    llm: {mock_llm_config.alias}
    default: true
"""
    )

    agents_dir = temp_dir / "agents"
    agents_dir.mkdir()
    (agents_dir / "new-agent.yml").write_text(
        f"""
name: new-agent
prompt: "new prompt"
llm: {mock_llm_config.alias}
default: false
"""
    )

    llm_config = BatchLLMConfig(llms=[mock_llm_config])
    config = await BatchAgentConfig.from_yaml(
        file_path=config_file, dir_path=agents_dir, batch_llm_config=llm_config
    )
    assert len(config.agents) == 2
    agent_names = {agent.name for agent in config.agents}
    assert "old-agent" in agent_names
    assert "new-agent" in agent_names


@pytest.mark.asyncio
async def test_update_agent_llm_directory_format(temp_dir):
    agents_dir = temp_dir / "agents"
    agents_dir.mkdir()

    (agents_dir / "test-agent.yml").write_text(
        """
name: test-agent
prompt: "test prompt"
llm: old-llm
default: true
"""
    )

    await BatchAgentConfig.update_agent_llm(
        file_path=temp_dir / "config.agents.yml",
        agent_name="test-agent",
        new_llm_name="new-llm",
        dir_path=agents_dir,
    )

    updated_content = (agents_dir / "test-agent.yml").read_text()
    assert "llm: new-llm" in updated_content
    assert "llm: old-llm" not in updated_content


@pytest.mark.asyncio
async def test_update_agent_llm_fallback_to_file(temp_dir):
    config_file = temp_dir / "config.agents.yml"
    config_file.write_text(
        """
agents:
  - name: test-agent
    prompt: "test prompt"
    llm: old-llm
    default: true
"""
    )

    await BatchAgentConfig.update_agent_llm(
        file_path=config_file,
        agent_name="test-agent",
        new_llm_name="new-llm",
        dir_path=temp_dir / "agents",
    )

    updated_content = config_file.read_text()
    assert "llm: new-llm" in updated_content
    assert "llm: old-llm" not in updated_content


@pytest.mark.asyncio
async def test_update_default_agent_directory_format(temp_dir):
    agents_dir = temp_dir / "agents"
    agents_dir.mkdir()

    (agents_dir / "agent1.yml").write_text(
        """
name: agent1
prompt: "test prompt"
llm: test-llm
default: true
"""
    )

    (agents_dir / "agent2.yml").write_text(
        """
name: agent2
prompt: "test prompt"
llm: test-llm
default: false
"""
    )

    await BatchAgentConfig.update_default_agent(
        file_path=temp_dir / "config.agents.yml",
        agent_name="agent2",
        dir_path=agents_dir,
    )

    agent1_content = (agents_dir / "agent1.yml").read_text()
    agent2_content = (agents_dir / "agent2.yml").read_text()

    assert "default: false" in agent1_content or "default: False" in agent1_content
    assert "default: true" in agent2_content or "default: True" in agent2_content



"""Integration tests for todo tools."""

from pathlib import Path

import pytest
from langchain_core.messages import AIMessage, HumanMessage

from src.agents.state import Todo
from src.tools.internal.todo import read_todos, write_todos


@pytest.mark.asyncio
async def test_todo_workflow(create_test_graph, temp_dir: Path):
    """Test write and read todos through the graph."""
    app = create_test_graph([write_todos, read_todos])

    # Write todos
    todos = [
        Todo(content="Task 1", status="pending"),
        Todo(content="Task 2", status="in_progress"),
    ]

    initial_state = {
        "messages": [
            HumanMessage(content="Write todos"),
            AIMessage(
                content="",
                tool_calls=[
                    {
                        "id": "call_1",
                        "name": "write_todos",
                        "args": {"todos": todos},
                    }
                ],
            ),
        ],
    }

    result = await app.ainvoke(
        initial_state,
        config={
            "configurable": {
                "thread_id": "test",
                "working_dir": str(temp_dir),
                "approval_mode": "aggressive",
            }
        },
    )

    # Verify todos were written to state
    assert result["todos"] is not None
    assert len(result["todos"]) == 2
    assert result["todos"][0]["content"] == "Task 1"



"""Tests for resume handler."""

from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from langgraph.checkpoint.base import CheckpointTuple

from src.cli.handlers.resume import ResumeHandler


@pytest.fixture
def sample_thread():
    """Create a sample thread dict for resume testing."""
    return {
        "thread_id": "thread-1",
        "last_message": "message 1",
        "timestamp": "",
    }


class TestResumeHandler:
    """Tests for ResumeHandler class."""

    @pytest.mark.asyncio
    @patch("src.cli.handlers.resume.initializer.get_threads")
    async def test_handle_with_no_threads(self, mock_get_threads, mock_session):
        """Test that handle shows error when no threads found."""
        handler = ResumeHandler(mock_session)
        mock_get_threads.return_value = []

        await handler.handle()

        mock_get_threads.assert_called_once_with(
            mock_session.context.agent, mock_session.context.working_dir
        )

    @pytest.mark.asyncio
    @patch("src.cli.handlers.resume.initializer.get_threads")
    async def test_handle_filters_current_thread(self, mock_get_threads, mock_session):
        """Test that handle filters out current thread from list."""
        handler = ResumeHandler(mock_session)

        threads = [
            {"thread_id": mock_session.context.thread_id, "last_message": "current"},
            {"thread_id": "other-thread", "last_message": "other"},
        ]
        mock_get_threads.return_value = threads

        with patch.object(handler, "_get_thread_selection", return_value=""):
            await handler.handle()

            mock_get_threads.assert_called_once()

    @pytest.mark.asyncio
    @patch("src.cli.handlers.resume.initializer.get_threads")
    async def test_handle_loads_selected_thread(self, mock_get_threads, mock_session):
        """Test that handle loads selected thread."""
        handler = ResumeHandler(mock_session)

        threads = [{"thread_id": "thread-1", "last_message": "message 1"}]
        mock_get_threads.return_value = threads

        with (
            patch.object(handler, "_get_thread_selection", return_value="thread-1"),
            patch.object(handler, "_load_thread", new_callable=AsyncMock) as mock_load,
        ):
            await handler.handle()

            mock_load.assert_called_once_with("thread-1", render_history=True)

    @pytest.mark.asyncio
    async def test_handle_with_direct_thread_id(self, mock_session):
        """Test that handle directly loads thread when thread_id provided."""
        handler = ResumeHandler(mock_session)

        with patch.object(handler, "_load_thread", new_callable=AsyncMock) as mock_load:
            await handler.handle(thread_id="specific-thread")

            mock_load.assert_called_once_with("specific-thread", render_history=True)

    @pytest.mark.asyncio
    @patch("src.cli.handlers.resume.Application")
    async def test_get_thread_selection_with_empty_list(
        self, mock_app_cls, mock_session
    ):
        """Test that _get_thread_selection returns empty string for no threads."""
        handler = ResumeHandler(mock_session)

        result = await handler._get_thread_selection([])

        assert result == ""
        mock_app_cls.assert_not_called()

    @pytest.mark.asyncio
    @patch("src.cli.handlers.resume.Application")
    async def test_get_thread_selection_with_selection(
        self, mock_app_cls, mock_session
    ):
        """Test that _get_thread_selection returns selected thread ID."""
        handler = ResumeHandler(mock_session)

        threads = [
            {"thread_id": "thread-1", "last_message": "message 1", "timestamp": ""},
            {"thread_id": "thread-2", "last_message": "message 2", "timestamp": ""},
        ]

        mock_app = AsyncMock()
        mock_app.run_async = AsyncMock()
        mock_app_cls.return_value = mock_app

        with patch("src.cli.handlers.resume.sys.stdout"):
            await handler._get_thread_selection(threads)
            mock_app.run_async.assert_called_once()

    @pytest.mark.asyncio
    @patch("src.cli.handlers.resume.Application")
    async def test_get_thread_selection_keyboard_interrupt(
        self, mock_app_cls, mock_session
    ):
        """Test that _get_thread_selection handles KeyboardInterrupt."""
        handler = ResumeHandler(mock_session)

        threads = [{"thread_id": "thread-1", "last_message": "msg", "timestamp": ""}]

        mock_app = AsyncMock()
        mock_app.run_async = AsyncMock(side_effect=KeyboardInterrupt())
        mock_app_cls.return_value = mock_app

        result = await handler._get_thread_selection(threads)

        assert result == ""

    @pytest.mark.asyncio
    @patch("src.cli.handlers.resume.initializer.get_checkpointer")
    async def test_load_thread_with_no_checkpoint(
        self, mock_get_checkpointer, mock_session, mock_checkpointer
    ):
        """Test that _load_thread handles missing checkpoint."""
        handler = ResumeHandler(mock_session)

        mock_checkpointer.aget_tuple.return_value = None
        mock_get_checkpointer.return_value.__aenter__.return_value = mock_checkpointer

        await handler._load_thread("thread-1")

        mock_checkpointer.aget_tuple.assert_called_once()

    @pytest.mark.asyncio
    @patch("src.cli.handlers.resume.initializer.get_checkpointer")
    async def test_load_thread_with_valid_checkpoint(
        self,
        mock_get_checkpointer,
        mock_session,
        mock_checkpointer,
        mock_checkpointer_tuple,
    ):
        """Test that _load_thread loads and renders messages."""
        handler = ResumeHandler(mock_session)

        mock_message = MagicMock()
        mock_message.id = "msg-1"

        checkpoint = mock_checkpointer_tuple.checkpoint.copy()
        checkpoint["channel_values"] = {
            "messages": [mock_message],
            "current_input_tokens": 100,
            "current_output_tokens": 50,
            "total_cost": 0.01,
        }
        checkpoint_tuple = CheckpointTuple(
            config=mock_checkpointer_tuple.config,
            checkpoint=checkpoint,
            metadata=mock_checkpointer_tuple.metadata,
            parent_config=mock_checkpointer_tuple.parent_config,
            pending_writes=mock_checkpointer_tuple.pending_writes,
        )

        mock_checkpointer.aget_tuple.return_value = checkpoint_tuple
        mock_get_checkpointer.return_value.__aenter__.return_value = mock_checkpointer

        await handler._load_thread("thread-1")

        mock_session.renderer.render_message.assert_called()
        mock_session.update_context.assert_called_once()

    @pytest.mark.asyncio
    @patch("src.cli.handlers.resume.initializer.get_checkpointer")
    async def test_load_thread_handles_exception(
        self, mock_get_checkpointer, mock_session
    ):
        """Test that _load_thread handles exceptions gracefully."""
        handler = ResumeHandler(mock_session)

        mock_get_checkpointer.side_effect = Exception("Test error")

        await handler._load_thread("thread-1")

    def test_format_thread_list_formats_correctly(self):
        """Test that _format_thread_list formats threads correctly."""
        threads = [
            {"thread_id": "t1", "last_message": "Hello world", "timestamp": ""},
            {"thread_id": "t2", "last_message": "Test message", "timestamp": ""},
        ]

        formatted = ResumeHandler._format_thread_list(threads, 0, 0, 5)

        assert formatted is not None

    def test_format_thread_list_with_scrolling(self):
        """Test that _format_thread_list handles scrolling window."""
        threads = [
            {"thread_id": f"t{i}", "last_message": f"msg {i}", "timestamp": ""}
            for i in range(10)
        ]

        formatted = ResumeHandler._format_thread_list(threads, 5, 3, 5)

        assert formatted is not None

    @pytest.mark.asyncio
    @patch("src.cli.handlers.resume.initializer.get_threads")
    async def test_handle_with_exception(self, mock_get_threads, mock_session):
        """Test that handle handles exceptions gracefully."""
        handler = ResumeHandler(mock_session)
        mock_get_threads.side_effect = Exception("Test error")

        await handler.handle()

        mock_get_threads.assert_called_once()



"""Tests for performance timer module."""

from unittest.mock import patch

import pytest

from src.cli.bootstrap.timer import _Timer, enable_timer, timer


@pytest.fixture
def mock_console():
    """Mock console for timer tests."""
    with patch("src.cli.bootstrap.timer.console.console") as mock_console:
        yield mock_console


@pytest.fixture
def mock_perf_counter():
    """Mock perf_counter that returns 0.0, then 1.5 seconds."""
    with patch(
        "src.cli.bootstrap.timer.time.perf_counter", side_effect=[0.0, 1.5]
    ) as mock:
        yield mock


@pytest.fixture
def timer_module_state():
    """Save and restore timer module's _enabled state."""
    import src.cli.bootstrap.timer as timer_module

    initial_state = timer_module._enabled
    yield timer_module
    timer_module._enabled = initial_state


class TestTimer:
    """Tests for _Timer class."""

    def test_timer_measures_elapsed_time(self, mock_console, mock_perf_counter):
        """Test that _Timer measures elapsed time."""
        timer_instance = _Timer("Test phase", enabled=True)

        with timer_instance:
            pass

        mock_console.print.assert_called_once()
        call_args = mock_console.print.call_args[0][0]
        assert "Test phase" in call_args
        assert "1.500s" in call_args

    def test_timer_does_nothing_when_disabled(self, mock_console):
        """Test that _Timer does nothing when disabled."""
        timer_instance = _Timer("Test phase", enabled=False)

        with timer_instance:
            pass

        mock_console.print.assert_not_called()

    def test_timer_enter_returns_self(self):
        """Test that __enter__ returns self."""
        timer_instance = _Timer("Test phase", enabled=False)

        result = timer_instance.__enter__()

        assert result is timer_instance

    @patch("src.cli.bootstrap.timer.time.perf_counter", side_effect=[0.0, 1.0])
    def test_timer_handles_exception_in_context(
        self,
        _mock_perf_counter,
        mock_console,
    ):
        """Test that _Timer handles exceptions in context."""
        timer_instance = _Timer("Test phase", enabled=True)

        try:
            with timer_instance:
                raise ValueError("Test error")
        except ValueError:
            pass

        mock_console.print.assert_called_once()

    @patch(
        "src.cli.bootstrap.timer.time.perf_counter",
        side_effect=[0.0, 0.123456],
    )
    def test_timer_formats_time_with_three_decimals(
        self,
        _mock_perf_counter,
        mock_console,
    ):
        """Test that _Timer formats time with three decimal places."""
        timer_instance = _Timer("Test phase", enabled=True)

        with timer_instance:
            pass

        call_args = mock_console.print.call_args[0][0]
        assert "0.123s" in call_args

    def test_timer_stores_phase_name(self):
        """Test that _Timer stores phase name."""
        timer_instance = _Timer("My Phase", enabled=True)

        assert timer_instance.phase_name == "My Phase"

    def test_timer_stores_enabled_flag(self):
        """Test that _Timer stores enabled flag."""
        timer_enabled = _Timer("Phase", enabled=True)
        timer_disabled = _Timer("Phase", enabled=False)

        assert timer_enabled.enabled is True
        assert timer_disabled.enabled is False


class TestEnableTimer:
    """Tests for enable_timer function."""

    def test_enable_timer_sets_global_flag(self, timer_module_state):
        """Test that enable_timer sets global _enabled flag."""
        enable_timer()

        assert timer_module_state._enabled is True

    def test_enable_timer_affects_timer_function(self, timer_module_state):
        """Test that enable_timer affects timer() function behavior."""
        timer_module_state._enabled = False
        timer_disabled = timer("Test")
        assert timer_disabled.enabled is False

        enable_timer()
        timer_enabled = timer("Test")
        assert timer_enabled.enabled is True


class TestTimerFunction:
    """Tests for timer function."""

    def test_timer_function_returns_timer_instance(self):
        """Test that timer() returns _Timer instance."""
        result = timer("Test phase")

        assert isinstance(result, _Timer)

    def test_timer_function_passes_phase_name(self):
        """Test that timer() passes phase name to _Timer."""
        result = timer("My Phase Name")

        assert result.phase_name == "My Phase Name"

    def test_timer_function_respects_global_enabled_flag(self, timer_module_state):
        """Test that timer() respects global _enabled flag."""
        timer_module_state._enabled = False
        result_disabled = timer("Test")
        assert result_disabled.enabled is False

        timer_module_state._enabled = True
        result_enabled = timer("Test")
        assert result_enabled.enabled is True

    def test_timer_function_can_be_used_as_context_manager(self, timer_module_state):
        """Test that timer() can be used as context manager."""
        timer_module_state._enabled = False

        with timer("Test phase"):
            pass

    @patch(
        "src.cli.bootstrap.timer.time.perf_counter",
        side_effect=[0.0, 2.5],
    )
    def test_timer_function_prints_when_enabled(
        self,
        _mock_perf_counter,
        timer_module_state,
        mock_console,
    ):
        """Test that timer() prints timing info when enabled."""
        timer_module_state._enabled = True

        with timer("Load config"):
            pass

        mock_console.print.assert_called_once()
        call_args = mock_console.print.call_args[0][0]
        assert "Load config" in call_args
        assert "2.500s" in call_args

    def test_timer_function_does_not_print_when_disabled(
        self, timer_module_state, mock_console
    ):
        """Test that timer() does not print when disabled."""
        timer_module_state._enabled = False

        with timer("Load config"):
            pass

        mock_console.print.assert_not_called()

    @patch(
        "src.cli.bootstrap.timer.time.perf_counter",
        side_effect=[0.0, 1.0, 2.0, 3.5],
    )
    def test_timer_function_multiple_timers_independent(
        self,
        _mock_perf_counter,
        timer_module_state,
        mock_console,
    ):
        """Test that multiple timers are independent."""
        timer_module_state._enabled = True

        with timer("Phase 1"):
            pass

        with timer("Phase 2"):
            pass

        assert mock_console.print.call_count == 2

        first_call = mock_console.print.call_args_list[0][0][0]
        second_call = mock_console.print.call_args_list[1][0][0]

        assert "Phase 1" in first_call
        assert "1.000s" in first_call
        assert "Phase 2" in second_call
        assert "1.500s" in second_call



"""Tests for ApprovalMiddleware."""

import tempfile
from pathlib import Path
from unittest.mock import AsyncMock, Mock, patch

import pytest
from langchain.tools.tool_node import ToolCallRequest
from langchain_core.messages import ToolMessage

from src.agents.context import AgentContext
from src.core.config import ApprovalMode, ToolApprovalConfig, ToolApprovalRule
from src.middleware.approval import (
    DENY,
    ApprovalMiddleware,
    create_field_extractor,
    create_field_transformer,
)


class TestApprovalMiddleware:
    """Tests for ApprovalMiddleware class."""

    def test_check_approval_rules_with_always_deny(self):
        """Test that always_deny rules are checked first."""
        config = ToolApprovalConfig(
            always_deny=[ToolApprovalRule(name="test_tool", args={"query": "bad"})],
            always_allow=[],
        )

        result = ApprovalMiddleware._check_approval_rules(
            config, "test_tool", {"query": "bad"}
        )

        assert result is False

    def test_check_approval_rules_with_always_allow(self):
        """Test that always_allow rules work."""
        config = ToolApprovalConfig(
            always_deny=[],
            always_allow=[ToolApprovalRule(name="test_tool", args={"query": "good"})],
        )

        result = ApprovalMiddleware._check_approval_rules(
            config, "test_tool", {"query": "good"}
        )

        assert result is True

    def test_check_approval_rules_with_no_match(self):
        """Test that None is returned when no rules match."""
        config = ToolApprovalConfig(
            always_deny=[],
            always_allow=[],
        )

        result = ApprovalMiddleware._check_approval_rules(
            config, "test_tool", {"query": "anything"}
        )

        assert result is None

    def test_check_approval_rules_deny_takes_precedence(self):
        """Test that deny rules take precedence over allow rules."""
        config = ToolApprovalConfig(
            always_deny=[ToolApprovalRule(name="test_tool", args=None)],
            always_allow=[ToolApprovalRule(name="test_tool", args=None)],
        )

        result = ApprovalMiddleware._check_approval_rules(config, "test_tool", {})

        assert result is False

    def test_check_approval_mode_bypass_semi_active(self):
        """Test that SEMI_ACTIVE mode never bypasses approval."""
        config = ToolApprovalConfig(always_deny=[], always_allow=[])

        result = ApprovalMiddleware._check_approval_mode_bypass(
            ApprovalMode.SEMI_ACTIVE, config, "test_tool", {}
        )

        assert result is False

    def test_check_approval_mode_bypass_active(self):
        """Test that ACTIVE mode bypasses unless explicitly denied."""
        config = ToolApprovalConfig(always_deny=[], always_allow=[])

        result = ApprovalMiddleware._check_approval_mode_bypass(
            ApprovalMode.ACTIVE, config, "test_tool", {}
        )

        assert result is True

    def test_check_approval_mode_bypass_active_with_deny(self):
        """Test that ACTIVE mode respects deny rules."""
        config = ToolApprovalConfig(
            always_deny=[ToolApprovalRule(name="test_tool", args=None)],
            always_allow=[],
        )

        result = ApprovalMiddleware._check_approval_mode_bypass(
            ApprovalMode.ACTIVE, config, "test_tool", {}
        )

        assert result is False

    def test_check_approval_mode_bypass_aggressive(self):
        """Test that AGGRESSIVE mode always bypasses approval."""
        config = ToolApprovalConfig(
            always_deny=[ToolApprovalRule(name="test_tool", args=None)],
            always_allow=[],
        )

        result = ApprovalMiddleware._check_approval_mode_bypass(
            ApprovalMode.AGGRESSIVE, config, "test_tool", {}
        )

        assert result is True

    def test_save_approval_decision_allow(self):
        """Test saving an allow decision."""
        with tempfile.TemporaryDirectory() as tmpdir:
            config_file = Path(tmpdir) / "config.approval.json"
            config = ToolApprovalConfig(always_deny=[], always_allow=[])
            config.save_to_json_file(config_file)

            ApprovalMiddleware._save_approval_decision(
                config, config_file, "test_tool", {"query": "test"}, allow=True
            )

            # Reload config
            loaded_config = ToolApprovalConfig.from_json_file(config_file)
            assert len(loaded_config.always_allow) == 1
            assert loaded_config.always_allow[0].name == "test_tool"
            assert loaded_config.always_allow[0].args == {"query": "test"}

    def test_save_approval_decision_deny(self):
        """Test saving a deny decision."""
        with tempfile.TemporaryDirectory() as tmpdir:
            config_file = Path(tmpdir) / "config.approval.json"
            config = ToolApprovalConfig(always_deny=[], always_allow=[])
            config.save_to_json_file(config_file)

            ApprovalMiddleware._save_approval_decision(
                config, config_file, "test_tool", {"query": "test"}, allow=False
            )

            loaded_config = ToolApprovalConfig.from_json_file(config_file)
            assert len(loaded_config.always_deny) == 1
            assert loaded_config.always_deny[0].name == "test_tool"

    def test_save_approval_decision_replaces_existing(self):
        """Test that saving a decision replaces existing rules."""
        with tempfile.TemporaryDirectory() as tmpdir:
            config_file = Path(tmpdir) / "config.approval.json"
            config = ToolApprovalConfig(
                always_allow=[
                    ToolApprovalRule(name="test_tool", args={"query": "test"})
                ],
                always_deny=[],
            )
            config.save_to_json_file(config_file)

            # Save a deny decision for the same tool/args
            ApprovalMiddleware._save_approval_decision(
                config, config_file, "test_tool", {"query": "test"}, allow=False
            )

            loaded_config = ToolApprovalConfig.from_json_file(config_file)
            assert len(loaded_config.always_allow) == 0
            assert len(loaded_config.always_deny) == 1

    @pytest.mark.asyncio
    async def test_awrap_tool_call_with_allow(self, create_mock_tool):
        """Test that tool is executed when allowed."""
        middleware = ApprovalMiddleware()

        # Mock request
        mock_tool = create_mock_tool("test_tool")
        request = Mock(spec=ToolCallRequest)
        request.tool_call = {
            "id": "call_1",
            "name": "test_tool",
            "args": {"query": "test"},
        }
        request.tool = mock_tool
        request.runtime = Mock()
        request.runtime.context = AgentContext(
            approval_mode=ApprovalMode.AGGRESSIVE,
            working_dir=Path("/tmp"),
        )

        # Mock handler
        handler = AsyncMock(
            return_value=ToolMessage(
                name="test_tool",
                content="result",
                tool_call_id="call_1",
            )
        )

        result = await middleware.awrap_tool_call(request, handler)

        assert isinstance(result, ToolMessage)
        assert result.content == "result"
        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_awrap_tool_call_with_deny(self, create_mock_tool):
        """Test that tool is not executed when denied."""
        with tempfile.TemporaryDirectory() as tmpdir:
            middleware = ApprovalMiddleware()

            # Create config with deny rule
            config_file = Path(tmpdir) / "config.approval.json"
            config = ToolApprovalConfig(
                always_deny=[ToolApprovalRule(name="test_tool", args=None)],
                always_allow=[],
            )
            config.save_to_json_file(config_file)

            mock_tool = create_mock_tool("test_tool")
            request = Mock(spec=ToolCallRequest)
            request.tool_call = {
                "id": "call_1",
                "name": "test_tool",
                "args": {"query": "test"},
            }
            request.tool = mock_tool
            request.runtime = Mock()
            request.runtime.context = AgentContext(
                approval_mode=ApprovalMode.SEMI_ACTIVE,
                working_dir=Path(tmpdir),
            )

            # Mock the interrupt to return DENY
            with patch("src.middleware.approval.interrupt", return_value=DENY):
                handler = AsyncMock()
                result = await middleware.awrap_tool_call(request, handler)

            assert isinstance(result, ToolMessage)
            assert getattr(result, "is_error", False) is True
            content = result.text
            assert "denied" in content.lower()
            handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_awrap_tool_call_with_exception(self, create_mock_tool):
        """Test that exceptions are caught and returned as error messages."""
        middleware = ApprovalMiddleware()

        mock_tool = create_mock_tool("test_tool")
        request = Mock(spec=ToolCallRequest)
        request.tool_call = {"id": "call_1", "name": "test_tool", "args": {}}
        request.tool = mock_tool
        request.runtime = Mock()
        request.runtime.context = AgentContext(
            approval_mode=ApprovalMode.AGGRESSIVE,
            working_dir=Path("/tmp"),
        )

        # Mock handler to raise exception
        handler = AsyncMock(side_effect=Exception("Test error"))

        result = await middleware.awrap_tool_call(request, handler)

        assert isinstance(result, ToolMessage)
        assert getattr(result, "is_error", False) is True
        content = result.text
        assert "Test error" in content


class TestFieldExtractor:
    """Tests for create_field_extractor helper."""

    def test_extract_single_field(self):
        """Test extracting a pattern from a single field."""
        extractor = create_field_extractor(
            {
                "command": r"(?P<command>\S+)",  # First word only
            }
        )

        result = extractor({"command": "echo hello world"})

        assert result["command"] == "echo"

    def test_extract_multiple_fields(self):
        """Test extracting patterns from multiple fields."""
        extractor = create_field_extractor(
            {
                "command": r"(?P<base_cmd>\S+)",
                "path": r"(?P<filename>[^/]+)$",
            }
        )

        result = extractor(
            {"command": "git commit -m 'message'", "path": "/home/user/file.txt"}
        )

        assert result["base_cmd"] == "git"
        assert result["filename"] == "file.txt"

    def test_no_match_keeps_original(self):
        """Test that fields without matches keep original values."""
        extractor = create_field_extractor(
            {
                "command": r"(?P<command>XYZ)",  # Won't match
            }
        )

        result = extractor({"command": "echo hello"})

        assert result["command"] == "echo hello"


class TestFieldTransformer:
    """Tests for create_field_transformer helper."""

    def test_transform_single_field(self):
        """Test transforming a single field."""
        transformer = create_field_transformer(
            {
                "command": lambda x: x.split()[0],
            }
        )

        result = transformer({"command": "echo hello world"})

        assert result["command"] == "echo"

    def test_transform_multiple_fields(self):
        """Test transforming multiple fields."""
        transformer = create_field_transformer(
            {
                "command": lambda x: x.upper(),
                "path": lambda x: x.replace("/", "_"),
            }
        )

        result = transformer({"command": "test", "path": "/home/user"})

        assert result["command"] == "TEST"
        assert result["path"] == "_home_user"

    def test_transform_handles_exceptions(self):
        """Test that transformation exceptions don't break the function."""
        transformer = create_field_transformer(
            {
                "command": lambda x: x.split()[10],  # Will raise IndexError
            }
        )

        result = transformer({"command": "echo"})

        # Should keep original value when transform fails
        assert result["command"] == "echo"



"""Image-related test fixtures."""

import base64
from pathlib import Path

import pytest


@pytest.fixture
def create_test_image(temp_dir):
    """Create a minimal test image.

    Args:
        temp_dir: Temporary directory fixture

    Returns:
        Callable that creates test images with various formats
    """

    def _create_image(filename: str, extension: str = ".png") -> Path:
        """Create a test image file.

        Args:
            filename: Name for the image file (without extension)
            extension: File extension (default: .png)

        Returns:
            Path to the created image file
        """
        # Minimal 1x1 PNG image (67 bytes)
        png_data = base64.b64decode(
            "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mNk+M9QDwADhgGAWjR9awAAAABJRU5ErkJggg=="
        )

        # Minimal 1x1 JPEG image
        jpeg_data = base64.b64decode(
            "/9j/4AAQSkZJRgABAQEAYABgAAD/2wBDAAgGBgcGBQgHBwcJCQgKDBQNDAsLDBkSEw8UHRofHh0aHBwgJC4nICIsIxwcKDcpLDAxNDQ0Hyc5PTgyPC4zNDL/wAALCAABAAEBAREA/8QAFAABAAAAAAAAAAAAAAAAAAAACv/EABQQAQAAAAAAAAAAAAAAAAAAAAD/2gAIAQEAAD8AKp//2Q=="
        )

        # Minimal 1x1 GIF image
        gif_data = base64.b64decode(
            "R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7"
        )

        # Minimal 1x1 WebP image
        webp_data = base64.b64decode(
            "UklGRiQAAABXRUJQVlA4IBgAAAAwAQCdASoBAAEAAwA0JaQAA3AA/vuUAAA="
        )

        image_data_map = {
            ".png": png_data,
            ".jpg": jpeg_data,
            ".jpeg": jpeg_data,
            ".gif": gif_data,
            ".webp": webp_data,
        }

        file_path = temp_dir / f"{filename}{extension}"
        file_path.write_bytes(image_data_map.get(extension, png_data))
        return file_path

    return _create_image



"""Tests for reference completer."""

import pytest
from prompt_toolkit.completion import CompleteEvent
from prompt_toolkit.document import Document

from src.cli.completers.reference import ReferenceCompleter, parse_reference
from src.cli.resolvers import RefType


class TestParseReference:
    """Tests for parse_reference function."""

    def test_parse_reference_with_file_type(self):
        """Test parsing reference with file type."""
        ref_type, value = parse_reference("@:file:path/to/file.txt")

        assert ref_type == RefType.FILE
        assert value == "path/to/file.txt"

    def test_parse_reference_without_at_symbol(self):
        """Test parsing reference without @ symbol."""
        ref_type, value = parse_reference(":file:path/to/file.txt")

        assert ref_type == RefType.FILE
        assert value == "path/to/file.txt"

    def test_parse_reference_with_no_type_prefix(self):
        """Test parsing reference with no type prefix."""
        ref_type, value = parse_reference("path/to/file.txt")

        assert ref_type is None
        assert value == "path/to/file.txt"

    def test_parse_reference_with_invalid_type(self):
        """Test parsing reference with invalid type."""
        ref_type, value = parse_reference(":invalid:path/to/file.txt")

        assert ref_type is None
        assert value == ":invalid:path/to/file.txt"

    def test_parse_reference_empty_value(self):
        """Test parsing reference with empty value."""
        ref_type, value = parse_reference(":file:")

        assert ref_type is None
        assert value == ":file:"

    def test_parse_reference_only_type(self):
        """Test parsing reference with only type prefix."""
        ref_type, value = parse_reference(":file")

        assert ref_type is None
        assert value == ":file"

    def test_parse_reference_with_at_prefix(self):
        """Test parsing reference strips @ prefix."""
        ref_type, value = parse_reference("@:file:test.txt")

        assert ref_type == RefType.FILE
        assert value == "test.txt"


class TestReferenceCompleter:
    """Tests for ReferenceCompleter class."""

    def test_init_creates_completer(self, temp_dir):
        """Test that __init__ creates completer with resolvers."""
        completer = ReferenceCompleter(temp_dir, max_suggestions=5)

        assert completer.working_dir == temp_dir
        assert completer.max_suggestions == 5
        assert RefType.FILE in completer.resolvers

    def test_get_completions_sync_returns_empty(self, temp_dir):
        """Test that sync get_completions returns empty iterator."""
        completer = ReferenceCompleter(temp_dir)
        document = Document(text="@:file:", cursor_position=7)
        event = CompleteEvent()

        completions = list(completer.get_completions(document, event))

        assert len(completions) == 0

    @pytest.mark.asyncio
    async def test_get_completions_async_no_at_ref(self, temp_dir):
        """Test async completions with no @ reference."""
        completer = ReferenceCompleter(temp_dir)
        document = Document(text="regular text", cursor_position=12)
        event = CompleteEvent()

        completions = []
        async for completion in completer.get_completions_async(document, event):
            completions.append(completion)

        assert len(completions) == 0

    @pytest.mark.asyncio
    async def test_get_completions_async_with_at_ref(self, temp_dir):
        """Test async completions with @ reference."""
        test_file = temp_dir / "test.txt"
        test_file.write_text("content")

        completer = ReferenceCompleter(temp_dir)
        document = Document(text="@:file:te", cursor_position=9)
        event = CompleteEvent()

        completions = []
        async for completion in completer.get_completions_async(document, event):
            completions.append(completion)

        assert len(completions) > 0

    @pytest.mark.asyncio
    async def test_get_completions_async_with_partial_path(self, temp_dir):
        """Test async completions with partial file path."""
        subdir = temp_dir / "subdir"
        subdir.mkdir()
        test_file = subdir / "file.txt"
        test_file.write_text("content")

        completer = ReferenceCompleter(temp_dir)
        document = Document(text="@:file:sub", cursor_position=10)
        event = CompleteEvent()

        completions = []
        async for completion in completer.get_completions_async(document, event):
            completions.append(completion)

        assert len(completions) > 0

    @pytest.mark.asyncio
    async def test_get_completions_async_respects_max_suggestions(self, temp_dir):
        """Test that completions respect max_suggestions limit."""
        for i in range(20):
            test_file = temp_dir / f"file{i}.txt"
            test_file.write_text("content")

        completer = ReferenceCompleter(temp_dir, max_suggestions=5)
        document = Document(text="@:file:file", cursor_position=11)
        event = CompleteEvent()

        completions = []
        async for completion in completer.get_completions_async(document, event):
            completions.append(completion)

        assert len(completions) <= 5

    @pytest.mark.asyncio
    async def test_get_completions_async_without_type_filter(self, temp_dir):
        """Test async completions without explicit type filter."""
        test_file = temp_dir / "test.txt"
        test_file.write_text("content")

        completer = ReferenceCompleter(temp_dir)
        document = Document(text="@te", cursor_position=3)
        event = CompleteEvent()

        completions = []
        async for completion in completer.get_completions_async(document, event):
            completions.append(completion)

        assert len(completions) >= 0

    def test_find_at_ref_with_at_symbol(self):
        """Test finding @ reference in text."""
        match = ReferenceCompleter._find_at_ref("some text @:file:path")

        assert match is not None
        assert match.group(1) == ":file:path"

    def test_find_at_ref_at_end(self):
        """Test finding @ reference at end of text."""
        match = ReferenceCompleter._find_at_ref("text @")

        assert match is not None
        assert match.group(1) == ""

    def test_find_at_ref_no_match(self):
        """Test finding @ reference with no match."""
        match = ReferenceCompleter._find_at_ref("text without reference")

        assert match is None

    def test_find_at_ref_multiple_refs(self):
        """Test finding @ reference finds last one."""
        match = ReferenceCompleter._find_at_ref("@:file:first @:file:last")

        assert match is not None
        assert match.group(1) == ":file:last"

    def test_find_at_ref_with_space_after(self):
        """Test finding @ reference stops at whitespace."""
        match = ReferenceCompleter._find_at_ref("text @:file:path more")

        assert match is None



"""Image handling utilities for multimodal support."""

import base64
import mimetypes
from pathlib import Path

# Supported image formats for common multimodal models
SUPPORTED_IMAGE_EXTENSIONS = {".png", ".jpg", ".jpeg", ".gif", ".webp"}
SUPPORTED_IMAGE_MIMES = {"image/png", "image/jpeg", "image/gif", "image/webp"}


def is_image_file(path: Path) -> bool:
    """Check if file is an image (any format)."""
    mime_type, _ = mimetypes.guess_type(str(path))
    return mime_type is not None and mime_type.startswith("image/")


def is_supported_image(path: Path) -> bool:
    """Check if file is a supported image format.

    Args:
        path: Path to the image file

    Returns:
        True if the file has a supported image extension
    """
    return path.suffix.lower() in SUPPORTED_IMAGE_EXTENSIONS


def get_image_mime_type(path: Path) -> str | None:
    """Get MIME type for an image file.

    Args:
        path: Path to the image file

    Returns:
        MIME type string (e.g., 'image/jpeg') or None if cannot be determined
    """
    mime_type, _ = mimetypes.guess_type(str(path))

    if mime_type and mime_type in SUPPORTED_IMAGE_MIMES:
        return mime_type

    # Fallback mapping for common extensions
    extension_to_mime = {
        ".png": "image/png",
        ".jpg": "image/jpeg",
        ".jpeg": "image/jpeg",
        ".gif": "image/gif",
        ".webp": "image/webp",
    }

    return extension_to_mime.get(path.suffix.lower())


def read_image_as_base64(path: Path) -> str:
    """Read image file and encode as base64 string.

    Args:
        path: Path to the image file

    Returns:
        Base64-encoded string of the image data

    Raises:
        FileNotFoundError: If the image file doesn't exist
        OSError: If there's an error reading the file
    """
    if not path.exists():
        raise FileNotFoundError(f"Image file not found: {path}")

    if not path.is_file():
        raise ValueError(f"Path is not a file: {path}")

    with open(path, "rb") as f:
        image_data = f.read()

    return base64.b64encode(image_data).decode("utf-8")


def is_image_path(text: str) -> bool:
    """Check if a string looks like an absolute path to an image file.

    Args:
        text: String to check

    Returns:
        True if the string is an absolute path to an existing image file
    """
    try:
        path = Path(text.strip())
        return (
            path.is_absolute()
            and path.exists()
            and path.is_file()
            and is_image_file(path)
        )
    except (OSError, ValueError):
        return False



"""Message content builder."""

import re
from pathlib import Path
from typing import Any

from src.cli.resolvers import FileResolver, ImageResolver, RefType


class MessageContentBuilder:
    """Builds message content with multimodal support."""

    def __init__(self, working_dir: Path):
        """Initialize message content builder."""
        self.working_dir = working_dir
        self.resolvers = {
            RefType.FILE: FileResolver(),
            RefType.IMAGE: ImageResolver(),
        }

    def extract_references(self, text: str) -> dict[RefType, list[str]]:
        """Extract all typed references and standalone paths from text."""
        references: dict[RefType, list[str]] = {}

        ref_pattern = r"@:([\w]+):((?:[^\s?,!.;]|[.,!?;](?!\s|$))+)"
        for match in re.finditer(ref_pattern, text):
            type_str, value = match.groups()
            try:
                ref_type = RefType(type_str)
                if ref_type not in references:
                    references[ref_type] = []
                references[ref_type].append(value)
            except ValueError:
                continue

        for word in text.split():
            if word.startswith("@:"):
                continue
            # Strip common punctuation from end of word
            cleaned_word = word.rstrip(".,!?;:")
            for resolver in self.resolvers.values():
                if resolver.is_standalone_reference(cleaned_word):
                    ref_type = resolver.type
                    if ref_type not in references:
                        references[ref_type] = []
                    references[ref_type].append(cleaned_word)
                    break

        return references

    def build(
        self, text: str
    ) -> tuple[str | list[str | dict[str, Any]], dict[str, str]]:
        """Build message content with multimodal support.

        Raises:
            FileNotFoundError: If referenced file/image doesn't exist
            ValueError: If referenced file/image is invalid
        """
        references = self.extract_references(text)

        if not references:
            return text, {}

        ctx = {"working_dir": str(self.working_dir)}
        reference_mapping = {}
        text_content = text
        errors: list[str] = []

        for ref_type, paths in references.items():
            resolver = self.resolvers[ref_type]
            for path in paths:
                resolved = resolver.resolve(path, ctx)
                reference_mapping[path] = resolved

                ref_pattern = rf"@:{ref_type.value}:{re.escape(path)}"
                try:
                    if resolver.build_content_block(resolved):
                        text_content = re.sub(ref_pattern, "", text_content)
                    else:
                        text_content = re.sub(ref_pattern, resolved, text_content)
                except (FileNotFoundError, ValueError) as e:
                    errors.append(str(e))
                    text_content = re.sub(ref_pattern, "", text_content)

        if errors:
            raise ValueError("\n".join(errors))

        content_blocks: list[str | dict[str, Any]] = []

        if text_content.strip():
            content_blocks.append({"type": "text", "text": text_content.strip()})

        for ref_type, paths in references.items():
            resolver = self.resolvers[ref_type]
            for path in paths:
                resolved = reference_mapping[path]
                try:
                    if block := resolver.build_content_block(resolved):
                        content_blocks.append(block)
                except (FileNotFoundError, ValueError):
                    pass

        if len(content_blocks) == 1:
            first_block = content_blocks[0]
            if isinstance(first_block, dict) and first_block.get("type") == "text":
                return first_block["text"], reference_mapping

        return content_blocks, reference_mapping



"""Middleware for LangChain agents.

This module contains middleware implementations for customizing agent behavior.
"""

from src.middleware.approval import ApprovalMiddleware
from src.middleware.compress_tool_output import CompressToolOutputMiddleware
from src.middleware.return_direct import ReturnDirectMiddleware
from src.middleware.token_cost import TokenCostMiddleware

__all__ = [
    "ApprovalMiddleware",
    "ReturnDirectMiddleware",
    "CompressToolOutputMiddleware",
    "TokenCostMiddleware",
]



"""Async SQLite checkpointer implementation with message indexing."""

from collections.abc import AsyncIterator, Callable
from contextlib import asynccontextmanager

from langchain_core.messages import AnyMessage, BaseMessage
from langchain_core.runnables import RunnableConfig
from langgraph.checkpoint.base import Checkpoint, CheckpointMetadata, CheckpointTuple
from langgraph.checkpoint.sqlite.aio import AsyncSqliteSaver

from src.checkpointer.base import BaseCheckpointer, HumanMessageEntry
from src.core.logging import get_logger

logger = get_logger(__name__)


class IndexedAsyncSqliteSaver(AsyncSqliteSaver, BaseCheckpointer):
    """AsyncSqliteSaver with automatic message indexing for performance."""

    async def setup(self) -> None:
        """Initialize tables including message index table."""
        await super().setup()

        async with self.lock:
            await self.conn.executescript(
                """
            CREATE TABLE IF NOT EXISTS checkpoint_messages (
                thread_id TEXT NOT NULL,
                checkpoint_id TEXT NOT NULL,
                checkpoint_ns TEXT NOT NULL DEFAULT '',
                message_idx INTEGER NOT NULL,
                message_type TEXT NOT NULL,
                message_preview TEXT,
                PRIMARY KEY (thread_id, checkpoint_id, message_idx)
            );
            CREATE INDEX IF NOT EXISTS idx_thread_ns_messages
                ON checkpoint_messages(thread_id, checkpoint_ns, message_type);
            CREATE INDEX IF NOT EXISTS idx_thread_lookup
                ON checkpoints(thread_id, checkpoint_ns);
            CREATE INDEX IF NOT EXISTS idx_checkpoint_id
                ON checkpoints(checkpoint_id);
            """
            )
            await self.conn.commit()

        logger.debug("Message index tables and indices created")

    async def aput(
        self,
        config: RunnableConfig,
        checkpoint: Checkpoint,
        metadata: CheckpointMetadata,
        new_versions: dict[str, int | float | str],
    ) -> RunnableConfig:
        """Save checkpoint and update message index."""
        result = await super().aput(config, checkpoint, metadata, new_versions)
        await self._index_messages(config, checkpoint)
        return result

    async def _index_messages(
        self, config: RunnableConfig, checkpoint: Checkpoint
    ) -> None:
        """Extract and index messages from checkpoint."""
        try:
            thread_id = config["configurable"].get("thread_id")
            checkpoint_id = checkpoint.get("id")
            checkpoint_ns = config["configurable"].get("checkpoint_ns", "")

            if not thread_id or not checkpoint_id:
                return

            messages = checkpoint.get("channel_values", {}).get("messages", [])
            if not messages:
                return

            async with self.lock:
                await self.conn.execute(
                    "DELETE FROM checkpoint_messages WHERE thread_id = ? AND checkpoint_id = ?",
                    (thread_id, checkpoint_id),
                )

                rows = [
                    (
                        thread_id,
                        checkpoint_id,
                        checkpoint_ns,
                        idx,
                        msg.type,
                        self._get_message_preview(msg),
                    )
                    for idx, msg in enumerate(messages)
                ]

                await self.conn.executemany(
                    """
                    INSERT OR REPLACE INTO checkpoint_messages
                    (thread_id, checkpoint_id, checkpoint_ns, message_idx, message_type, message_preview)
                    VALUES (?, ?, ?, ?, ?, ?)
                    """,
                    rows,
                )

                await self.conn.commit()

            logger.debug(
                f"Indexed {len(messages)} messages for thread {thread_id}, checkpoint {checkpoint_id}"
            )

        except Exception as e:
            logger.warning(f"Failed to index messages: {e}")

    async def get_threads(self) -> set[str]:
        """Get all thread IDs."""
        try:
            async with self.lock:
                cursor = await self.conn.execute(
                    "SELECT DISTINCT thread_id FROM checkpoints WHERE checkpoint_ns = ''"
                )
                rows = await cursor.fetchall()

            return {row[0] for row in rows if row[0]}

        except Exception as e:
            logger.error(f"Failed to get thread IDs: {e}")
            return set()

    async def get_human_messages(
        self,
        thread_id: str,
        latest: CheckpointTuple,
        on_indexing: Callable[[], None] | None = None,
    ) -> tuple[list[HumanMessageEntry], list[BaseMessage]]:
        """Get human messages with replay metadata."""
        if not latest or not latest.checkpoint:
            return [], []

        all_messages = latest.checkpoint.get("channel_values", {}).get("messages", [])
        channel_values = latest.checkpoint.get("channel_values", {})
        checkpoint_ns = latest.config.get("configurable", {}).get("checkpoint_ns", "")

        # Check if indexed
        async with self.lock:
            cursor = await self.conn.execute(
                "SELECT COUNT(*) FROM checkpoint_messages WHERE thread_id = ? AND checkpoint_ns = ?",
                (thread_id, checkpoint_ns),
            )
            row = await cursor.fetchone()
            has_index = row and row[0] > 0

        # Lazy index if needed
        if not has_index:
            if on_indexing:
                on_indexing()

            logger.debug(f"Indexing thread {thread_id} namespace {checkpoint_ns!r}...")
            history = await self.get_history(latest)

            for checkpoint_tuple in history:
                checkpoint = checkpoint_tuple.checkpoint
                if checkpoint:
                    config = RunnableConfig(
                        configurable={
                            "thread_id": checkpoint_tuple.config.get(
                                "configurable", {}
                            ).get("thread_id"),
                            "checkpoint_ns": checkpoint_tuple.config.get(
                                "configurable", {}
                            ).get("checkpoint_ns", ""),
                        }
                    )
                    await self._index_messages(config, checkpoint)

            logger.debug(
                f"Indexing complete for {thread_id} namespace {checkpoint_ns!r}"
            )

        # Build mapping of message index -> checkpoint_id using the index
        # Query: for each unique checkpoint, get the max message_idx (= message count)
        # Filter by namespace to only get checkpoints from the current branch
        checkpoint_by_msg_count = {}
        async with self.lock:
            cursor = await self.conn.execute(
                """
                SELECT checkpoint_id, MAX(message_idx) + 1 as msg_count
                FROM checkpoint_messages
                WHERE thread_id = ? AND checkpoint_ns = ?
                GROUP BY checkpoint_id
                ORDER BY msg_count
                """,
                (thread_id, checkpoint_ns),
            )
            rows = await cursor.fetchall()

        for checkpoint_id, msg_count in rows:
            checkpoint_by_msg_count[msg_count] = checkpoint_id

        human_messages = []
        for i, msg in enumerate(all_messages):
            if msg.type == "human":
                human_messages.append(
                    HumanMessageEntry(
                        text=getattr(msg, "short_content", None) or msg.text,
                        reference_mapping=msg.additional_kwargs.get(
                            "reference_mapping", {}
                        ),
                        messages_before_count=i,
                        checkpoint_id=checkpoint_by_msg_count.get(i),
                        input_tokens=channel_values.get("current_input_tokens"),
                        output_tokens=channel_values.get("current_output_tokens"),
                        total_cost=channel_values.get("total_cost"),
                    )
                )

        return human_messages, all_messages

    async def get_history(self, latest: CheckpointTuple) -> list[CheckpointTuple]:
        """Get checkpoint history in chronological order (oldest first)."""
        history = []
        current: CheckpointTuple | None = latest

        while current is not None:
            history.append(current)
            current = (
                await self.aget_tuple(current.parent_config)
                if current.parent_config
                else None
            )

        history.reverse()
        return history

    async def delete_after(self, thread_id: str, checkpoint_id: str | None) -> int:
        """Delete checkpoints after checkpoint_id. Returns count deleted."""
        # Get checkpoint IDs to delete
        if checkpoint_id is None:
            checkpoints_to_delete = await self._get_all_checkpoint_ids(thread_id)
        else:
            checkpoints_to_delete = await self._get_checkpoints_after(
                thread_id, checkpoint_id
            )

        if not checkpoints_to_delete:
            return 0

        # Delete them
        async with self.lock:
            placeholders = ",".join(["?"] * len(checkpoints_to_delete))
            await self.conn.execute(
                f"DELETE FROM checkpoints WHERE thread_id = ? AND checkpoint_id IN ({placeholders})",
                (thread_id, *checkpoints_to_delete),
            )
            await self.conn.execute(
                f"DELETE FROM writes WHERE thread_id = ? AND checkpoint_id IN ({placeholders})",
                (thread_id, *checkpoints_to_delete),
            )
            await self.conn.execute(
                f"DELETE FROM checkpoint_messages WHERE thread_id = ? AND checkpoint_id IN ({placeholders})",
                (thread_id, *checkpoints_to_delete),
            )
            await self.conn.commit()

        return len(checkpoints_to_delete)

    async def _get_all_checkpoint_ids(self, thread_id: str) -> list[str]:
        """Get all checkpoint IDs for a thread."""
        async with self.lock:
            cursor = await self.conn.execute(
                "SELECT checkpoint_id FROM checkpoints WHERE thread_id = ?",
                (thread_id,),
            )
            rows = await cursor.fetchall()
        return [row[0] for row in rows]

    async def _get_checkpoints_after(
        self, thread_id: str, checkpoint_id: str
    ) -> list[str]:
        """Get checkpoint IDs after checkpoint_id. Uses index if available."""
        # Try index-based lookup
        async with self.lock:
            cursor = await self.conn.execute(
                "SELECT MAX(message_idx) + 1, checkpoint_ns FROM checkpoint_messages WHERE thread_id = ? AND checkpoint_id = ?",
                (thread_id, checkpoint_id),
            )
            row = await cursor.fetchone()

        if row and row[0] is not None:
            # Fast path: use index, filter by namespace to stay in same branch
            target_msg_count = row[0]
            checkpoint_ns = row[1]
            async with self.lock:
                cursor = await self.conn.execute(
                    """
                    SELECT DISTINCT checkpoint_id FROM checkpoint_messages
                    WHERE thread_id = ? AND checkpoint_ns = ?
                    GROUP BY checkpoint_id
                    HAVING MAX(message_idx) + 1 > ?
                    """,
                    (thread_id, checkpoint_ns, target_msg_count),
                )
                rows = await cursor.fetchall()
            return [row[0] for row in rows]

        # Slow path: walk history
        config = RunnableConfig(configurable={"thread_id": thread_id})
        latest = await self.aget_tuple(config)
        if not latest:
            return []

        history = await self.get_history(latest)
        idx = next(
            (
                i
                for i, cp in enumerate(history)
                if cp.checkpoint.get("id") == checkpoint_id
            ),
            None,
        )
        if idx is None:
            return []

        return [
            cp_id for cp in history[idx + 1 :] if (cp_id := cp.checkpoint.get("id"))
        ]

    @staticmethod
    def _get_message_preview(msg: AnyMessage) -> str:
        """Get truncated message preview."""
        try:
            text = getattr(msg, "short_content", None) or getattr(msg, "text", "")
            if isinstance(text, list):
                text = " ".join(str(item) for item in text)
            text = str(text).replace("\n", " ")
            return text[:100]
        except Exception:
            return ""


class AsyncSqliteCheckpointer:
    """Wrapper for IndexedAsyncSqliteSaver with connection management."""

    @staticmethod
    @asynccontextmanager
    async def create(
        connection_string: str,
    ) -> AsyncIterator[IndexedAsyncSqliteSaver]:
        """Create an async SQLite checkpointer with indexing and connection management.

        Args:
            connection_string: SQLite database file path or ":memory:" for in-memory

        Yields:
            IndexedAsyncSqliteSaver: The configured checkpointer instance
        """
        logger.debug(
            f"Creating indexed SQLite checkpointer with connection: {connection_string}"
        )

        try:
            async with IndexedAsyncSqliteSaver.from_conn_string(
                connection_string
            ) as saver:
                await saver.setup()
                logger.debug("Indexed SQLite checkpointer created successfully")
                yield saver  # type: ignore[misc]
        except Exception as e:
            logger.error(f"Failed to create SQLite checkpointer: {e}")
            raise



"""In-memory checkpointer implementation."""

from collections.abc import Callable

from langchain_core.messages import BaseMessage
from langchain_core.runnables import RunnableConfig
from langgraph.checkpoint.base import CheckpointTuple
from langgraph.checkpoint.memory import MemorySaver

from src.checkpointer.base import BaseCheckpointer, HumanMessageEntry
from src.core.logging import get_logger

logger = get_logger(__name__)


class MemoryCheckpointer(MemorySaver, BaseCheckpointer):
    """In-memory checkpointer that extends LangGraph's MemorySaver.

    This implementation stores checkpoints in memory and does not persist
    across application restarts. Useful for development and testing.
    """

    def __init__(self):
        """Initialize the memory checkpointer."""
        super().__init__()
        logger.debug("Memory checkpointer initialized")

    def delete_checkpoints(
        self, thread_id: str, checkpoint_ns: str, checkpoint_ids: list[str]
    ) -> int:
        """Delete specific checkpoints from memory storage.

        Args:
            thread_id: Thread ID
            checkpoint_ns: Checkpoint namespace
            checkpoint_ids: List of checkpoint IDs to delete

        Returns:
            Number of checkpoints deleted
        """
        deleted_count = 0

        # MemorySaver stores checkpoints as: storage[thread_id][checkpoint_ns][checkpoint_id]
        if thread_id in self.storage and checkpoint_ns in self.storage[thread_id]:
            for checkpoint_id in checkpoint_ids:
                if checkpoint_id in self.storage[thread_id][checkpoint_ns]:
                    del self.storage[thread_id][checkpoint_ns][checkpoint_id]
                    deleted_count += 1
                    logger.debug(
                        f"Deleted checkpoint {checkpoint_id} from thread {thread_id}"
                    )

        # Also delete from writes: writes[(thread_id, checkpoint_ns, checkpoint_id)]
        for checkpoint_id in checkpoint_ids:
            write_key = (thread_id, checkpoint_ns, checkpoint_id)
            if write_key in self.writes:
                del self.writes[write_key]

        return deleted_count

    async def get_threads(self) -> set[str]:
        """Get all thread IDs."""
        return set(self.storage.keys())

    async def get_history(self, latest: CheckpointTuple) -> list[CheckpointTuple]:
        """Get checkpoint history in chronological order (oldest first)."""
        history = []
        current: CheckpointTuple | None = latest

        while current is not None:
            history.append(current)
            current = (
                await self.aget_tuple(current.parent_config)
                if current.parent_config
                else None
            )

        history.reverse()
        return history

    async def delete_after(self, thread_id: str, checkpoint_id: str | None) -> int:
        """Delete all checkpoints after checkpoint_id across all namespaces."""
        config = RunnableConfig(configurable={"thread_id": thread_id})
        latest = await self.aget_tuple(config)
        if not latest:
            return 0

        history = await self.get_history(latest)

        idx = (
            -1
            if checkpoint_id is None
            else next(
                (
                    i
                    for i, cp in enumerate(history)
                    if cp.checkpoint.get("id") == checkpoint_id
                ),
                None,
            )
        )
        if idx is None:
            return 0

        to_delete = history[idx + 1 :]
        if not to_delete:
            return 0

        checkpoint_ids_in_history = {
            cp_id for cp in to_delete if (cp_id := cp.checkpoint.get("id"))
        }

        total_deleted = 0
        if thread_id in self.storage:
            for ns in list(self.storage[thread_id].keys()):
                ns_checkpoints = list(self.storage[thread_id][ns].keys())

                to_delete_in_ns = [
                    cp_id
                    for cp_id in ns_checkpoints
                    if cp_id in checkpoint_ids_in_history or checkpoint_id is None
                ]

                if to_delete_in_ns or checkpoint_id is None:
                    if checkpoint_id is None:
                        to_delete_in_ns = ns_checkpoints

                    deleted = self.delete_checkpoints(thread_id, ns, to_delete_in_ns)
                    total_deleted += deleted

        return total_deleted

    async def get_human_messages(
        self,
        thread_id: str,
        latest: CheckpointTuple,
        on_indexing: Callable[[], None] | None = None,
    ) -> tuple[list[HumanMessageEntry], list[BaseMessage]]:
        """Get human messages with replay metadata."""
        if not latest or not latest.checkpoint:
            return [], []

        all_messages = latest.checkpoint.get("channel_values", {}).get("messages", [])
        channel_values = latest.checkpoint.get("channel_values", {})

        checkpoint_history = await self.get_history(latest)
        checkpoint_by_msg_count = {}
        for checkpoint_tuple in checkpoint_history:
            checkpoint = checkpoint_tuple.checkpoint
            if checkpoint and "channel_values" in checkpoint:
                msg_count = len(checkpoint["channel_values"].get("messages", []))
                checkpoint_by_msg_count[msg_count] = checkpoint.get("id")

        human_messages = []
        for i, msg in enumerate(all_messages):
            if msg.type == "human":
                human_messages.append(
                    HumanMessageEntry(
                        text=getattr(msg, "short_content", None) or msg.text,
                        reference_mapping=msg.additional_kwargs.get(
                            "reference_mapping", {}
                        ),
                        messages_before_count=i,
                        checkpoint_id=checkpoint_by_msg_count.get(i),
                        input_tokens=channel_values.get("current_input_tokens"),
                        output_tokens=channel_values.get("current_output_tokens"),
                        total_cost=channel_values.get("total_cost"),
                    )
                )

        return human_messages, all_messages



"""Checkpointer-related test fixtures."""

import uuid
from unittest.mock import AsyncMock

import pytest
from langchain_core.runnables import RunnableConfig
from langgraph.checkpoint.base import CheckpointTuple

from src.core.config import CheckpointerConfig, CheckpointerProvider


@pytest.fixture
def mock_checkpointer_tuple():
    """Create a mock CheckpointTuple for testing."""
    return CheckpointTuple(
        config=RunnableConfig(configurable={"thread_id": str(uuid.uuid4())}),
        checkpoint={
            "v": 1,
            "channel_values": {
                "messages": [],
                "current_input_tokens": 0,
                "current_output_tokens": 0,
                "total_cost": 0.0,
            },
            "channel_versions": {},
            "versions_seen": {},
            "updated_channels": [],
            "ts": "2024-01-01T00:00:00.000Z",
            "id": str(uuid.uuid4()),
        },
        metadata={},
        parent_config=None,
        pending_writes=None,
    )


@pytest.fixture
def mock_checkpointer(mock_checkpointer_tuple):
    """Create a mock LangGraph checkpointer with CheckpointTuple for integration tests."""
    checkpointer = AsyncMock()
    checkpointer.aget_tuple = AsyncMock(return_value=mock_checkpointer_tuple)
    checkpointer.alist = AsyncMock(return_value=[mock_checkpointer_tuple])
    checkpointer.aput = AsyncMock()

    return checkpointer


@pytest.fixture
def mock_checkpointer_config():
    """Create a mock CheckpointerConfig for testing."""
    return CheckpointerConfig(type=CheckpointerProvider.MEMORY)



"""Tests for model handler."""

from unittest.mock import AsyncMock, patch

import pytest

from src.cli.handlers.models import ModelHandler
from src.core.config import AgentConfig


class TestModelHandler:
    """Tests for ModelHandler class."""

    @pytest.mark.asyncio
    @patch("src.cli.handlers.models.initializer.load_llms_config")
    @patch("src.cli.handlers.models.initializer.load_agent_config")
    async def test_handle_with_no_other_models(
        self,
        mock_load_agent,
        mock_load_llms,
        mock_session,
        mock_agent_config,
        mock_llms_config,
    ):
        """Test that handle shows error when no other models available."""
        handler = ModelHandler(mock_session)
        mock_load_agent.return_value = mock_agent_config

        mock_llms_config.llms = [mock_agent_config.llm]
        mock_load_llms.return_value = mock_llms_config

        with (
            patch.object(
                handler,
                "_get_agent_selection",
                return_value=("agent", "test-agent", mock_agent_config),
            ),
            patch("src.cli.handlers.models.console") as mock_console,
        ):
            await handler.handle()

            mock_load_agent.assert_called_once()
            mock_console.print_error.assert_called_once()

    @pytest.mark.asyncio
    @patch("src.cli.handlers.models.initializer.update_agent_llm")
    @patch("src.cli.handlers.models.initializer.load_llms_config")
    @patch("src.cli.handlers.models.initializer.load_agent_config")
    async def test_handle_updates_agent_model(
        self,
        mock_load_agent,
        mock_load_llms,
        mock_update_llm,
        mock_session,
        mock_agent_config,
        mock_llm_config,
        mock_llms_config,
    ):
        """Test that handle updates agent model successfully."""
        handler = ModelHandler(mock_session)
        mock_load_agent.return_value = mock_agent_config

        alt_llm_config = mock_llm_config.model_copy(
            update={"alias": "alt-model", "model": "claude-3-opus-20240229"}
        )

        mock_llms_config.llms = [mock_agent_config.llm, alt_llm_config]
        mock_load_llms.return_value = mock_llms_config

        with (
            patch.object(
                handler,
                "_get_agent_selection",
                return_value=("agent", "test-agent", mock_agent_config),
            ),
            patch.object(handler, "_get_model_selection", return_value="alt-model"),
        ):
            await handler.handle()

            mock_update_llm.assert_called_once_with(
                "test-agent", "alt-model", mock_session.context.working_dir
            )
            mock_session.update_context.assert_called_once()

    @pytest.mark.asyncio
    @patch("src.cli.handlers.models.initializer.update_subagent_llm")
    @patch("src.cli.handlers.models.initializer.load_llms_config")
    @patch("src.cli.handlers.models.initializer.load_agent_config")
    async def test_handle_updates_subagent_model(
        self,
        mock_load_agent,
        mock_load_llms,
        mock_update_subagent_llm,
        mock_session,
        mock_agent_config,
        mock_llm_config,
        mock_llms_config,
    ):
        """Test that handle updates subagent model successfully."""
        handler = ModelHandler(mock_session)

        subagent_llm = mock_llm_config.model_copy(
            update={"alias": "subagent-model", "model": "claude-3-haiku-20240307"}
        )
        subagent = AgentConfig(name="subagent-1", llm=subagent_llm, prompt="")
        mock_agent_config.subagents = [subagent]
        mock_load_agent.return_value = mock_agent_config

        alt_llm_config = mock_llm_config.model_copy(
            update={"alias": "alt-model", "model": "claude-3-opus-20240229"}
        )

        mock_llms_config.llms = [subagent_llm, alt_llm_config]
        mock_load_llms.return_value = mock_llms_config

        with (
            patch.object(
                handler,
                "_get_agent_selection",
                return_value=("subagent", "subagent-1", subagent),
            ),
            patch.object(handler, "_get_model_selection", return_value="alt-model"),
        ):
            await handler.handle()

            mock_update_subagent_llm.assert_called_once_with(
                "subagent-1", "alt-model", mock_session.context.working_dir
            )

    @pytest.mark.asyncio
    @patch("src.cli.handlers.models.Application")
    async def test_get_agent_selection_with_no_agents(self, mock_app_cls, mock_session):
        """Test that _get_agent_selection returns None for empty list."""
        handler = ModelHandler(mock_session)

        result = await handler._get_agent_selection([])

        assert result is None
        mock_app_cls.assert_not_called()

    @pytest.mark.asyncio
    @patch("src.cli.handlers.models.Application")
    async def test_get_agent_selection_with_selection(
        self, mock_app_cls, mock_session, mock_agent_config
    ):
        """Test that _get_agent_selection returns selected agent."""
        handler = ModelHandler(mock_session)

        agents = [("agent", "test-agent", mock_agent_config)]

        mock_app = AsyncMock()
        mock_app.run_async = AsyncMock()
        mock_app_cls.return_value = mock_app

        with patch("src.cli.handlers.models.sys.stdout"):
            await handler._get_agent_selection(agents)
            mock_app.run_async.assert_called_once()

    @pytest.mark.asyncio
    @patch("src.cli.handlers.models.Application")
    async def test_get_agent_selection_keyboard_interrupt(
        self, mock_app_cls, mock_session, mock_agent_config
    ):
        """Test that _get_agent_selection handles KeyboardInterrupt."""
        handler = ModelHandler(mock_session)

        agents = [("agent", "test-agent", mock_agent_config)]

        mock_app = AsyncMock()
        mock_app.run_async = AsyncMock(side_effect=KeyboardInterrupt())
        mock_app_cls.return_value = mock_app

        result = await handler._get_agent_selection(agents)

        assert result is None

    @pytest.mark.asyncio
    @patch("src.cli.handlers.models.Application")
    async def test_get_model_selection_with_no_models(self, mock_app_cls, mock_session):
        """Test that _get_model_selection returns empty string for no models."""
        handler = ModelHandler(mock_session)

        result = await handler._get_model_selection([], "test-model", "test-model")

        assert result == ""
        mock_app_cls.assert_not_called()

    @pytest.mark.asyncio
    @patch("src.cli.handlers.models.Application")
    async def test_get_model_selection_with_selection(
        self, mock_app_cls, mock_session, mock_llm_config
    ):
        """Test that _get_model_selection returns selected model."""
        handler = ModelHandler(mock_session)

        mock_app = AsyncMock()
        mock_app.run_async = AsyncMock()
        mock_app_cls.return_value = mock_app

        with patch("src.cli.handlers.models.sys.stdout"):
            await handler._get_model_selection(
                [mock_llm_config], "test-model", "test-model"
            )
            mock_app.run_async.assert_called_once()

    @pytest.mark.asyncio
    @patch("src.cli.handlers.models.Application")
    async def test_get_model_selection_keyboard_interrupt(
        self, mock_app_cls, mock_session, mock_llm_config
    ):
        """Test that _get_model_selection handles KeyboardInterrupt."""
        handler = ModelHandler(mock_session)

        mock_app = AsyncMock()
        mock_app.run_async = AsyncMock(side_effect=KeyboardInterrupt())
        mock_app_cls.return_value = mock_app

        result = await handler._get_model_selection(
            [mock_llm_config], "test-model", "test-model"
        )

        assert result == ""

    def test_format_agent_list_formats_correctly(self, mock_session, mock_agent_config):
        """Test that _format_agent_list formats agents correctly."""
        handler = ModelHandler(mock_session)
        agents = [("agent", "test-agent", mock_agent_config)]

        formatted = handler._format_agent_list(agents, 0)

        assert formatted is not None
        # Verify the formatted text contains the agent name and model
        formatted_str = "".join(str(item[1]) for item in formatted)
        assert "test-agent" in formatted_str
        # Should show context model for main agent
        assert mock_session.context.model in formatted_str

    def test_format_model_list_formats_correctly(self, mock_llm_config):
        """Test that _format_model_list formats models correctly."""
        formatted = ModelHandler._format_model_list(
            [mock_llm_config], 0, "test-model", "test-model"
        )

        assert formatted is not None
        # Verify the formatted text contains the model name and indicators
        formatted_str = "".join(str(item[1]) for item in formatted)
        assert mock_llm_config.alias in formatted_str

    @pytest.mark.asyncio
    @patch("src.cli.handlers.models.initializer.load_agent_config")
    async def test_handle_with_exception(self, mock_load_agent, mock_session):
        """Test that handle handles exceptions gracefully."""
        handler = ModelHandler(mock_session)
        mock_load_agent.side_effect = Exception("Test error")

        await handler.handle()

        mock_load_agent.assert_called_once()

    @pytest.mark.asyncio
    @patch("src.cli.handlers.models.initializer.load_llms_config")
    @patch("src.cli.handlers.models.initializer.load_agent_config")
    async def test_handle_cancelled_agent_selection(
        self, mock_load_agent, mock_load_llms, mock_session, mock_agent_config
    ):
        """Test that handle returns when agent selection is cancelled."""
        handler = ModelHandler(mock_session)
        mock_load_agent.return_value = mock_agent_config

        with patch.object(handler, "_get_agent_selection", return_value=None):
            await handler.handle()

            mock_load_llms.assert_not_called()

    @pytest.mark.asyncio
    @patch("src.cli.handlers.models.initializer.load_llms_config")
    @patch("src.cli.handlers.models.initializer.load_agent_config")
    async def test_handle_cancelled_model_selection(
        self,
        mock_load_agent,
        mock_load_llms,
        mock_session,
        mock_agent_config,
        mock_llm_config,
        mock_llms_config,
    ):
        """Test that handle returns when model selection is cancelled."""
        handler = ModelHandler(mock_session)
        mock_load_agent.return_value = mock_agent_config

        alt_llm_config = mock_llm_config.model_copy(
            update={"alias": "alt-model", "model": "claude-3-opus-20240229"}
        )

        mock_llms_config.llms = [mock_agent_config.llm, alt_llm_config]
        mock_load_llms.return_value = mock_llms_config

        with (
            patch.object(
                handler,
                "_get_agent_selection",
                return_value=("agent", "test-agent", mock_agent_config),
            ),
            patch.object(handler, "_get_model_selection", return_value=""),
        ):
            await handler.handle()

            mock_session.update_context.assert_not_called()



"""Tests for server command handler."""

import json
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from src.cli.bootstrap.server import (
    _set_assistant_version,
    _upsert_assistant,
    _wait_for_server_ready,
    generate_langgraph_json,
    get_graph,
    handle_server_command,
)
from src.core.constants import CONFIG_LANGGRAPH_FILE_NAME


@pytest.fixture
def mock_http_client():
    """Create a mock HTTP client for server tests."""
    mock_client = AsyncMock()
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.json = MagicMock(
        return_value={"assistant_id": "test-id", "version": 1, "name": "Test"}
    )
    mock_client.get = AsyncMock(return_value=mock_response)
    mock_client.post = AsyncMock(
        return_value=MagicMock(status_code=200, json=MagicMock(return_value=[]))
    )
    return mock_client


@pytest.fixture
def mock_subprocess():
    """Create a mock subprocess for server tests."""
    mock_process = MagicMock()
    mock_process.wait = MagicMock(return_value=0)
    return mock_process


@pytest.fixture
def patch_get_graph(mock_initializer):
    """Patch initializer for get_graph tests."""
    with patch("src.cli.bootstrap.server.initializer", mock_initializer):
        yield mock_initializer


@pytest.fixture
def patch_server_dependencies(
    mock_initializer,
    mock_agent_config,
    mock_llm_config,
    mock_http_client,
    mock_subprocess,
):
    """Patch dependencies for handle_server_command tests."""
    mock_agent_config.llm = mock_llm_config

    with (
        patch("src.cli.bootstrap.server.initializer", mock_initializer),
        patch(
            "src.cli.bootstrap.server.subprocess.Popen", return_value=mock_subprocess
        ) as mock_popen,
        patch("src.cli.bootstrap.server.httpx.AsyncClient") as mock_client_cls,
    ):
        mock_client_cls.return_value.__aenter__ = AsyncMock(
            return_value=mock_http_client
        )
        mock_client_cls.return_value.__aexit__ = AsyncMock()

        yield {
            "initializer": mock_initializer,
            "agent_config": mock_agent_config,
            "llm_config": mock_llm_config,
            "popen": mock_popen,
            "client_cls": mock_client_cls,
            "client": mock_http_client,
            "process": mock_subprocess,
        }


class TestGenerateLanggraphJson:
    """Tests for generate_langgraph_json function."""

    def test_generate_langgraph_json_creates_file(self, temp_dir):
        """Test that generate_langgraph_json creates langgraph.json."""
        generate_langgraph_json(temp_dir)

        langgraph_json_path = temp_dir / CONFIG_LANGGRAPH_FILE_NAME
        assert langgraph_json_path.exists()

    def test_generate_langgraph_json_has_correct_structure(self, temp_dir):
        """Test that generated langgraph.json has correct structure."""
        generate_langgraph_json(temp_dir)

        langgraph_json_path = temp_dir / CONFIG_LANGGRAPH_FILE_NAME
        with open(langgraph_json_path) as f:
            config = json.load(f)

        assert "dependencies" in config
        assert "graphs" in config
        assert "agent" in config["graphs"]
        assert config["graphs"]["agent"] == "src/cli/bootstrap/server.py:get_graph"

    def test_generate_langgraph_json_includes_env_when_exists(self, temp_dir):
        """Test that .env is included in config when it exists."""
        env_file = temp_dir / ".env"
        env_file.write_text("TEST=value")

        generate_langgraph_json(temp_dir)

        langgraph_json_path = temp_dir / CONFIG_LANGGRAPH_FILE_NAME
        with open(langgraph_json_path) as f:
            config = json.load(f)

        assert "env" in config
        assert config["env"] == ".env"

    def test_generate_langgraph_json_excludes_env_when_missing(self, temp_dir):
        """Test that .env is excluded from config when it doesn't exist."""
        generate_langgraph_json(temp_dir)

        langgraph_json_path = temp_dir / CONFIG_LANGGRAPH_FILE_NAME
        with open(langgraph_json_path) as f:
            config = json.load(f)

        assert "env" not in config

    def test_generate_langgraph_json_creates_parent_dir(self, temp_dir):
        """Test that parent directory is created if it doesn't exist."""
        config_dir = temp_dir / CONFIG_LANGGRAPH_FILE_NAME.parent
        assert not config_dir.exists()

        generate_langgraph_json(temp_dir)

        assert config_dir.exists()
        assert config_dir.is_dir()


class TestGetGraph:
    """Tests for get_graph function."""

    @pytest.mark.asyncio
    async def test_get_graph_reads_env_variables(
        self, temp_dir, monkeypatch, patch_get_graph, mock_graph
    ):
        """Test that get_graph reads environment variables."""
        monkeypatch.setenv("LANGREPL_AGENT", "test-agent")
        monkeypatch.setenv("LANGREPL_MODEL", "test-model")
        monkeypatch.setenv("LANGREPL_WORKING_DIR", str(temp_dir))

        result = await get_graph()

        assert result == mock_graph

    @pytest.mark.asyncio
    async def test_get_graph_requires_working_dir(self, monkeypatch):
        """Test that get_graph raises ValueError when LANGREPL_WORKING_DIR is missing."""
        monkeypatch.delenv("LANGREPL_WORKING_DIR", raising=False)

        with pytest.raises(ValueError, match="LANGREPL_WORKING_DIR"):
            await get_graph()

    @pytest.mark.asyncio
    async def test_get_graph_handles_missing_agent(
        self, temp_dir, monkeypatch, patch_get_graph, mock_graph
    ):
        """Test that get_graph handles missing LANGREPL_AGENT."""
        monkeypatch.delenv("LANGREPL_AGENT", raising=False)
        monkeypatch.setenv("LANGREPL_WORKING_DIR", str(temp_dir))

        result = await get_graph()

        assert result == mock_graph

    @pytest.mark.asyncio
    async def test_get_graph_handles_missing_model(
        self, temp_dir, monkeypatch, patch_get_graph, mock_graph
    ):
        """Test that get_graph handles missing LANGREPL_MODEL."""
        monkeypatch.setenv("LANGREPL_AGENT", "test-agent")
        monkeypatch.delenv("LANGREPL_MODEL", raising=False)
        monkeypatch.setenv("LANGREPL_WORKING_DIR", str(temp_dir))

        result = await get_graph()

        assert result == mock_graph


class TestWaitForServerReady:
    """Tests for _wait_for_server_ready function."""

    @pytest.mark.asyncio
    async def test_wait_for_server_ready_returns_true_on_success(self):
        """Test that _wait_for_server_ready returns True when server is ready."""
        mock_client = AsyncMock()
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_client.get = AsyncMock(return_value=mock_response)

        result = await _wait_for_server_ready(
            mock_client, "http://localhost:8000", timeout_seconds=1
        )

        assert result is True

    @pytest.mark.asyncio
    async def test_wait_for_server_ready_returns_false_on_timeout(self):
        """Test that _wait_for_server_ready returns False on timeout."""
        import httpx

        mock_client = AsyncMock()
        mock_client.get = AsyncMock(side_effect=httpx.HTTPError("Connection refused"))

        result = await _wait_for_server_ready(
            mock_client, "http://localhost:8000", timeout_seconds=1
        )

        assert result is False

    @pytest.mark.asyncio
    async def test_wait_for_server_ready_retries_on_failure(self):
        """Test that _wait_for_server_ready retries on failure."""
        mock_client = AsyncMock()
        mock_response_fail = MagicMock()
        mock_response_fail.status_code = 500
        mock_response_success = MagicMock()
        mock_response_success.status_code = 200

        mock_client.get = AsyncMock(
            side_effect=[mock_response_fail, mock_response_success]
        )

        result = await _wait_for_server_ready(
            mock_client, "http://localhost:8000", timeout_seconds=2
        )

        assert result is True
        assert mock_client.get.call_count == 2


class TestSetAssistantVersion:
    """Tests for _set_assistant_version function."""

    @pytest.mark.asyncio
    async def test_set_assistant_version_makes_post_request(self):
        """Test that _set_assistant_version makes POST request."""
        mock_client = AsyncMock()
        assistant = {"assistant_id": "test-id", "version": 1}

        await _set_assistant_version(mock_client, "http://localhost:8000", assistant)

        mock_client.post.assert_called_once()
        call_args = mock_client.post.call_args
        assert "test-id" in call_args[0][0]
        assert call_args[1]["json"]["version"] == 1


class TestUpsertAssistant:
    """Tests for _upsert_assistant function."""

    @pytest.mark.asyncio
    @patch("src.cli.bootstrap.server._set_assistant_version", new_callable=AsyncMock)
    async def test_upsert_assistant_creates_new_assistant(
        self,
        _mock_set_version,
    ):
        """Test that _upsert_assistant creates new assistant when none exists."""
        mock_client = AsyncMock()

        mock_search_response = MagicMock()
        mock_search_response.status_code = 200
        mock_search_response.json = MagicMock(return_value=[])

        mock_create_response = MagicMock()
        mock_create_response.status_code = 200
        mock_create_response.json = MagicMock(
            return_value={"assistant_id": "new-id", "version": 1}
        )

        post_calls = [mock_search_response, mock_create_response]

        async def mock_post(*args, **kwargs):
            return post_calls.pop(0)

        mock_client.post = mock_post

        assistant, was_updated = await _upsert_assistant(
            mock_client, "http://localhost:8000", "Test Assistant", {}
        )

        assert assistant is not None
        assert was_updated is False
        assert assistant["assistant_id"] == "new-id"

    @pytest.mark.asyncio
    async def test_upsert_assistant_updates_existing_assistant(self):
        """Test that _upsert_assistant updates existing assistant."""
        mock_client = AsyncMock()

        mock_search_response = MagicMock()
        mock_search_response.status_code = 200
        mock_search_response.json = MagicMock(
            return_value=[{"assistant_id": "existing-id"}]
        )

        mock_update_response = MagicMock()
        mock_update_response.status_code = 200
        mock_update_response.json = MagicMock(
            return_value={"assistant_id": "existing-id", "version": 2}
        )

        mock_client.post = AsyncMock(return_value=mock_search_response)
        mock_client.patch = AsyncMock(return_value=mock_update_response)

        assistant, was_updated = await _upsert_assistant(
            mock_client, "http://localhost:8000", "Test Assistant", {}
        )

        assert assistant is not None
        assert was_updated is True
        assert assistant["assistant_id"] == "existing-id"

    @pytest.mark.asyncio
    async def test_upsert_assistant_handles_failure(self):
        """Test that _upsert_assistant handles failure gracefully."""
        import httpx

        mock_client = AsyncMock()

        async def mock_post(*args, **kwargs):
            raise httpx.HTTPError("Network error")

        mock_client.post = mock_post

        assistant, was_updated = await _upsert_assistant(
            mock_client, "http://localhost:8000", "Test Assistant", {}
        )

        assert assistant is None
        assert was_updated is False


class TestHandleServerCommand:
    """Tests for handle_server_command function."""

    @pytest.mark.asyncio
    async def test_handle_server_command_generates_config(
        self, mock_app_args, patch_server_dependencies
    ):
        """Test that handle_server_command generates langgraph.json."""
        await handle_server_command(mock_app_args)

        langgraph_json_path = (
            Path(mock_app_args.working_dir) / CONFIG_LANGGRAPH_FILE_NAME
        )
        assert langgraph_json_path.exists()

    @pytest.mark.asyncio
    async def test_handle_server_command_starts_subprocess(
        self, mock_app_args, patch_server_dependencies
    ):
        """Test that handle_server_command starts langgraph dev subprocess."""
        await handle_server_command(mock_app_args)

        patch_server_dependencies["popen"].assert_called_once()
        call_args = patch_server_dependencies["popen"].call_args[0][0]
        assert any("langgraph" in str(arg) for arg in call_args)
        assert "dev" in call_args

    @pytest.mark.asyncio
    async def test_handle_server_command_sets_env_variables(
        self, mock_app_args, patch_server_dependencies
    ):
        """Test that handle_server_command sets correct environment variables."""
        mock_app_args.model = "test-model"

        await handle_server_command(mock_app_args)

        call_env = patch_server_dependencies["popen"].call_args[1]["env"]
        assert call_env["LANGREPL_WORKING_DIR"] == mock_app_args.working_dir
        assert call_env["LANGREPL_AGENT"] == "test-agent"
        assert call_env["LANGREPL_MODEL"] == "test-model"

    @pytest.mark.asyncio
    @patch(
        "src.cli.bootstrap.server._wait_for_server_ready",
        new_callable=AsyncMock,
        return_value=True,
    )
    async def test_handle_server_command_waits_for_server(
        self,
        mock_wait,
        mock_app_args,
        patch_server_dependencies,
    ):
        """Test that handle_server_command waits for server to be ready."""
        await handle_server_command(mock_app_args)

        mock_wait.assert_awaited_once()

    @pytest.mark.asyncio
    @patch(
        "src.cli.bootstrap.server._wait_for_server_ready",
        new_callable=AsyncMock,
        return_value=False,
    )
    async def test_handle_server_command_kills_on_timeout(
        self,
        mock_wait,
        mock_app_args,
        patch_server_dependencies,
    ):
        """Test that handle_server_command terminates process on timeout."""
        result = await handle_server_command(mock_app_args)

        mock_wait.assert_awaited_once()
        patch_server_dependencies["process"].terminate.assert_called_once()
        patch_server_dependencies["process"].wait.assert_called()
        assert result == 1

    @pytest.mark.asyncio
    async def test_handle_server_command_handles_exception(
        self, mock_app_args, patch_server_dependencies
    ):
        """Test that handle_server_command handles exceptions gracefully."""
        patch_server_dependencies["initializer"].load_agent_config.side_effect = (
            Exception("Config error")
        )

        result = await handle_server_command(mock_app_args)

        assert result == 1

    @pytest.mark.asyncio
    async def test_handle_server_command_adds_gitignore(
        self, mock_app_args, patch_server_dependencies
    ):
        """Test that handle_server_command adds .langgraph_api to gitignore."""
        git_dir = Path(mock_app_args.working_dir) / ".git" / "info"
        git_dir.mkdir(parents=True)

        await handle_server_command(mock_app_args)

        exclude_file = git_dir / "exclude"
        assert exclude_file.exists()
        content = exclude_file.read_text()
        assert ".langgraph_api/" in content



"""Integration tests for terminal tools."""

from pathlib import Path

import pytest
from langchain_core.messages import AIMessage, HumanMessage

from src.tools.impl.terminal import get_directory_structure, run_command


@pytest.mark.asyncio
async def test_run_command(create_test_graph, agent_context):
    """Test running a command through the graph."""
    app = create_test_graph([run_command])

    initial_state = {
        "messages": [
            HumanMessage(content="Run command"),
            AIMessage(
                content="",
                tool_calls=[
                    {
                        "id": "call_1",
                        "name": "run_command",
                        "args": {"command": "echo hello"},
                    }
                ],
            ),
        ],
    }

    result = await app.ainvoke(
        initial_state,
        config={"configurable": {"thread_id": "test"}},
        context=agent_context,
    )

    # Check that command output is in messages
    tool_messages = [m for m in result["messages"] if m.type == "tool"]
    assert tool_messages
    assert "hello" in tool_messages[0].content


@pytest.mark.asyncio
async def test_directory_structure(create_test_graph, agent_context, temp_dir: Path):
    """Test getting directory structure through the graph."""
    # Setup: create some files
    (temp_dir / "file1.txt").write_text("content")
    (temp_dir / "subdir").mkdir()

    app = create_test_graph([get_directory_structure])

    initial_state = {
        "messages": [
            HumanMessage(content="Get structure"),
            AIMessage(
                content="",
                tool_calls=[
                    {
                        "id": "call_1",
                        "name": "get_directory_structure",
                        "args": {"dir_path": "."},
                    }
                ],
            ),
        ],
    }

    result = await app.ainvoke(
        initial_state,
        config={"configurable": {"thread_id": "test"}},
        context=agent_context,
    )

    tool_messages = [m for m in result["messages"] if m.type == "tool"]
    assert tool_messages
    assert "file1.txt" in tool_messages[0].content


@pytest.mark.asyncio
async def test_run_command_failure(create_test_graph, agent_context):
    """Test running an invalid command through the graph."""
    app = create_test_graph([run_command])

    initial_state = {
        "messages": [
            HumanMessage(content="Run invalid command"),
            AIMessage(
                content="",
                tool_calls=[
                    {
                        "id": "call_1",
                        "name": "run_command",
                        "args": {"command": "nonexistent_command_xyz"},
                    }
                ],
            ),
        ],
    }

    result = await app.ainvoke(
        initial_state,
        config={"configurable": {"thread_id": "test"}},
        context=agent_context,
    )

    # Check that error is returned
    tool_messages = [m for m in result["messages"] if m.type == "tool"]
    assert tool_messages
    assert (
        "error" in tool_messages[0].content.lower()
        or "not found" in tool_messages[0].content.lower()
    )


@pytest.mark.asyncio
async def test_directory_structure_with_special_chars(
    create_test_graph, agent_context, temp_dir: Path
):
    """Test that directory paths with special characters are properly escaped."""
    # Create a directory with spaces and special characters in the name
    special_dir = temp_dir / "dir with spaces"
    special_dir.mkdir()
    (special_dir / "test.txt").write_text("content")

    app = create_test_graph([get_directory_structure])

    initial_state = {
        "messages": [
            HumanMessage(content="Get structure"),
            AIMessage(
                content="",
                tool_calls=[
                    {
                        "id": "call_1",
                        "name": "get_directory_structure",
                        "args": {"dir_path": str(special_dir)},
                    }
                ],
            ),
        ],
    }

    result = await app.ainvoke(
        initial_state,
        config={"configurable": {"thread_id": "test"}},
        context=agent_context,
    )

    # Should successfully handle the directory despite special characters
    tool_messages = [m for m in result["messages"] if m.type == "tool"]
    assert tool_messages
    assert "test.txt" in tool_messages[0].content



"""Tests for CLI session module."""

from contextlib import asynccontextmanager
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from src.cli.core.session import Session


class TestSessionInit:
    """Tests for Session initialization."""

    def test_init_creates_all_components(self, mock_context):
        """Test that __init__ creates all required components."""
        session = Session(mock_context)

        assert session.context == mock_context
        assert session.renderer is not None
        assert session.command_dispatcher is not None
        assert session.message_dispatcher is not None
        assert session.prompt is not None

    def test_init_sets_default_state(self, mock_context):
        """Test that __init__ sets default session state."""
        session = Session(mock_context)

        assert session.graph is None
        assert session.graph_context is None
        assert session.running is False
        assert session.needs_reload is False
        assert session.prefilled_text is None
        assert session.prefilled_reference_mapping == {}

    @patch("src.cli.core.session.InteractivePrompt")
    def test_init_registers_mode_change_callback(
        self,
        mock_prompt_cls,
        mock_context,
    ):
        """Test that __init__ registers approval mode change callback."""
        mock_prompt = MagicMock()
        mock_prompt_cls.return_value = mock_prompt

        Session(mock_context)

        mock_prompt.set_mode_change_callback.assert_called_once()
        callback = mock_prompt.set_mode_change_callback.call_args[0][0]
        assert callable(callback)


class TestSessionStart:
    """Tests for Session.start() method."""

    @pytest.mark.asyncio
    @patch.object(Session, "_main_loop", new_callable=AsyncMock)
    @patch("src.cli.core.session.initializer.get_graph")
    async def test_start_loads_graph(
        self,
        mock_get_graph,
        _mock_main_loop,
        mock_context,
    ):
        """Test that start() loads the graph."""
        graph_instance = MagicMock()

        @asynccontextmanager
        async def graph_context(*args, **kwargs):
            yield graph_instance

        mock_get_graph.side_effect = graph_context
        session = Session(mock_context)

        await session.start()

        assert session.graph is graph_instance

    @pytest.mark.asyncio
    @patch.object(Session, "_main_loop", new_callable=AsyncMock)
    @patch("src.cli.core.session.initializer.get_graph")
    async def test_start_shows_welcome_by_default(
        self,
        mock_get_graph,
        _mock_main_loop,
        mock_context,
    ):
        """Test that start() shows welcome message by default."""
        graph_instance = MagicMock()

        @asynccontextmanager
        async def graph_context(*args, **kwargs):
            yield graph_instance

        mock_get_graph.side_effect = graph_context
        session = Session(mock_context)
        session.renderer = MagicMock()

        await session.start()

        session.renderer.show_welcome.assert_called_once_with(mock_context)

    @pytest.mark.asyncio
    @patch.object(Session, "_main_loop", new_callable=AsyncMock)
    @patch("src.cli.core.session.initializer.get_graph")
    async def test_start_hides_welcome_when_requested(
        self,
        mock_get_graph,
        _mock_main_loop,
        mock_context,
    ):
        """Test that start() hides welcome message when show_welcome=False."""
        graph_instance = MagicMock()

        @asynccontextmanager
        async def graph_context(*args, **kwargs):
            yield graph_instance

        mock_get_graph.side_effect = graph_context
        session = Session(mock_context)
        session.renderer = MagicMock()

        await session.start(show_welcome=False)

        session.renderer.show_welcome.assert_not_called()

    @pytest.mark.asyncio
    @patch.object(Session, "_main_loop", new_callable=AsyncMock)
    @patch("src.cli.core.session.initializer.get_graph")
    async def test_start_calls_main_loop(
        self,
        mock_get_graph,
        mock_main_loop,
        mock_context,
    ):
        """Test that start() calls main loop."""
        graph_instance = MagicMock()

        @asynccontextmanager
        async def graph_context(*args, **kwargs):
            yield graph_instance

        mock_get_graph.side_effect = graph_context
        session = Session(mock_context)

        await session.start()

        mock_main_loop.assert_called_once()


class TestSessionUpdateContext:
    """Tests for Session.update_context() method."""

    def test_update_context_updates_single_field(self, mock_context):
        """Test that update_context() updates a single field."""
        session = Session(mock_context)
        original_value = mock_context.current_input_tokens

        session.update_context(current_input_tokens=1000)

        assert mock_context.current_input_tokens == 1000
        assert mock_context.current_input_tokens != original_value

    def test_update_context_updates_multiple_fields(self, mock_context):
        """Test that update_context() updates multiple fields at once."""
        session = Session(mock_context)

        session.update_context(
            current_input_tokens=1000, current_output_tokens=2000, total_cost=3.5
        )

        assert mock_context.current_input_tokens == 1000
        assert mock_context.current_output_tokens == 2000
        assert mock_context.total_cost == 3.5

    def test_update_context_with_agent_triggers_reload(self, mock_context):
        """Test that updating agent triggers reload."""
        session = Session(mock_context)
        session.running = True
        session.needs_reload = False

        session.update_context(agent="new-agent")

        assert session.needs_reload is True
        assert session.running is False

    def test_update_context_with_model_triggers_reload(self, mock_context):
        """Test that updating model triggers reload."""
        session = Session(mock_context)
        session.running = True
        session.needs_reload = False

        session.update_context(model="new-model")

        assert session.needs_reload is True
        assert session.running is False

    def test_update_context_with_both_agent_and_model_triggers_reload(
        self, mock_context
    ):
        """Test that updating both agent and model triggers reload."""
        session = Session(mock_context)
        session.running = True
        session.needs_reload = False

        session.update_context(agent="new-agent", model="new-model")

        assert session.needs_reload is True
        assert session.running is False

    def test_update_context_with_tokens_does_not_trigger_reload(self, mock_context):
        """Test that updating tokens doesn't trigger reload."""
        session = Session(mock_context)
        session.running = True
        session.needs_reload = False

        session.update_context(current_input_tokens=1000)

        assert session.needs_reload is False
        assert session.running is True

    def test_update_context_ignores_invalid_fields(self, mock_context):
        """Test that update_context() ignores fields not in context."""
        session = Session(mock_context)

        session.update_context(invalid_field="value")

        assert not hasattr(mock_context, "invalid_field")

    def test_update_context_with_thread_id(self, mock_context):
        """Test that update_context() can update thread_id."""
        session = Session(mock_context)
        new_thread_id = "new-thread-123"

        session.update_context(thread_id=new_thread_id)

        assert mock_context.thread_id == new_thread_id


class TestSessionHandleApprovalModeChange:
    """Tests for Session._handle_approval_mode_change() method."""

    def test_handle_approval_mode_change_cycles_mode(self, mock_context):
        """Test that _handle_approval_mode_change cycles approval mode."""
        session = Session(mock_context)
        original_mode = mock_context.approval_mode

        session._handle_approval_mode_change()

        assert mock_context.approval_mode != original_mode

    def test_handle_approval_mode_change_refreshes_prompt_style(self, mock_context):
        """Test that _handle_approval_mode_change refreshes prompt style."""
        session = Session(mock_context)
        session.prompt = MagicMock()

        session._handle_approval_mode_change()

        session.prompt.refresh_style.assert_called_once()



import pytest
import yaml

from src.core.config import (
    BatchAgentConfig,
    BatchCheckpointerConfig,
    BatchLLMConfig,
    ToolApprovalRule,
)


class TestToolApprovalRuleMatchesCall:
    def test_exact_name_match_no_args(self):
        rule = ToolApprovalRule(name="read_file", args=None)
        assert rule.matches_call("read_file", {}) is True
        assert rule.matches_call("read_file", {"path": "/tmp/file"}) is True

    def test_name_mismatch(self):
        rule = ToolApprovalRule(name="read_file", args=None)
        assert rule.matches_call("write_file", {}) is False

    def test_exact_args_match(self):
        rule = ToolApprovalRule(name="read_file", args={"path": "/tmp/test"})
        assert rule.matches_call("read_file", {"path": "/tmp/test"}) is True

    def test_args_mismatch(self):
        rule = ToolApprovalRule(name="read_file", args={"path": "/tmp/test"})
        assert rule.matches_call("read_file", {"path": "/tmp/other"}) is False

    def test_regex_pattern_match(self):
        rule = ToolApprovalRule(name="read_file", args={"path": r"/tmp/.*"})
        assert rule.matches_call("read_file", {"path": "/tmp/test"}) is True
        assert rule.matches_call("read_file", {"path": "/tmp/file.txt"}) is True

    def test_regex_pattern_no_match(self):
        rule = ToolApprovalRule(name="read_file", args={"path": r"/tmp/.*"})
        assert rule.matches_call("read_file", {"path": "/home/test"}) is False

    def test_missing_required_arg(self):
        rule = ToolApprovalRule(name="read_file", args={"path": "/tmp/test"})
        assert rule.matches_call("read_file", {}) is False

    def test_multiple_args_all_match(self):
        rule = ToolApprovalRule(
            name="copy_file", args={"src": "/tmp/.*", "dst": "/backup/.*"}
        )
        assert (
            rule.matches_call("copy_file", {"src": "/tmp/file", "dst": "/backup/file"})
            is True
        )

    def test_multiple_args_partial_match(self):
        rule = ToolApprovalRule(
            name="copy_file", args={"src": "/tmp/.*", "dst": "/backup/.*"}
        )
        assert (
            rule.matches_call("copy_file", {"src": "/tmp/file", "dst": "/home/file"})
            is False
        )


class TestBatchAgentConfigGetDefaultAgent:
    def test_explicit_default(self, mock_agent_config):
        agent1 = mock_agent_config.model_copy(
            update={"name": "agent1", "default": False}
        )
        agent2 = mock_agent_config.model_copy(
            update={"name": "agent2", "default": True}
        )

        config = BatchAgentConfig(agents=[agent1, agent2])
        default_agent = config.get_default_agent()
        assert default_agent is not None
        assert default_agent.name == "agent2"

    def test_no_explicit_default_returns_first(self, mock_agent_config):
        agent1 = mock_agent_config.model_copy(
            update={"name": "agent1", "default": False}
        )
        agent2 = mock_agent_config.model_copy(
            update={"name": "agent2", "default": False}
        )

        config = BatchAgentConfig(agents=[agent1, agent2])
        default_agent = config.get_default_agent()
        assert default_agent is not None
        assert default_agent.name == "agent1"

    def test_empty_agents_returns_none(self):
        config = BatchAgentConfig(agents=[])
        assert config.get_default_agent() is None

    def test_get_agent_by_name(self, mock_agent_config):
        agent1 = mock_agent_config.model_copy(update={"name": "agent1"})
        agent2 = mock_agent_config.model_copy(update={"name": "agent2"})

        config = BatchAgentConfig(agents=[agent1, agent2])
        agent2 = config.get_agent_config("agent2")
        assert agent2 is not None
        assert agent2.name == "agent2"
        assert config.get_agent_config("nonexistent") is None


class TestBatchAgentConfigValidation:
    def test_multiple_defaults_raises_error(self, mock_agent_config):
        agent1 = mock_agent_config.model_copy(
            update={"name": "agent1", "default": True}
        )
        agent2 = mock_agent_config.model_copy(
            update={"name": "agent2", "default": True}
        )

        with pytest.raises(ValueError, match="Multiple agents marked as default"):
            BatchAgentConfig(agents=[agent1, agent2])

    def test_single_default_is_valid(self, mock_agent_config):
        agent1 = mock_agent_config.model_copy(
            update={"name": "agent1", "default": True}
        )

        config = BatchAgentConfig(agents=[agent1])
        default_agent = config.get_default_agent()
        assert default_agent is not None
        assert default_agent.name == "agent1"


class TestDuplicateValidation:
    """Test duplicate detection in configs."""

    @pytest.mark.asyncio
    async def test_duplicate_agents(self, temp_dir, mock_llm_config):
        """Test that duplicate agent names raise an error."""
        file_path = temp_dir / "config.agents.yml"
        file_path.write_text(
            yaml.dump(
                {
                    "agents": [
                        {"name": "my-agent", "default": True, "llm": "test-model"},
                        {"name": "my-agent", "llm": "test-model"},  # Duplicate
                    ]
                }
            )
        )

        batch_llm = BatchLLMConfig(llms=[mock_llm_config])

        with pytest.raises(ValueError, match=r"Duplicate agent 'name': 'my-agent'"):
            await BatchAgentConfig.from_yaml(
                file_path=file_path, batch_llm_config=batch_llm
            )

    @pytest.mark.asyncio
    async def test_duplicate_llms(self, temp_dir):
        """Test that duplicate LLM aliases raise an error."""
        file_path = temp_dir / "config.llms.yml"
        file_path.write_text(
            yaml.dump(
                {
                    "llms": [
                        {
                            "alias": "model",
                            "provider": "anthropic",
                            "model": "claude-3-5-sonnet-20241022",
                            "max_tokens": 4096,
                            "temperature": 0.7,
                        },
                        {
                            "alias": "model",  # Duplicate
                            "provider": "openai",
                            "model": "gpt-4",
                            "max_tokens": 8192,
                            "temperature": 0.5,
                        },
                    ]
                }
            )
        )

        with pytest.raises(ValueError, match=r"Duplicate llm 'alias': 'model'"):
            await BatchLLMConfig.from_yaml(file_path=file_path)

    @pytest.mark.asyncio
    async def test_duplicate_checkpointers(self, temp_dir):
        """Test that duplicate checkpointer types raise an error."""
        file_path = temp_dir / "config.checkpointers.yml"
        file_path.write_text(
            yaml.dump(
                {
                    "checkpointers": [
                        {"type": "sqlite", "max_connections": 10},
                        {"type": "sqlite", "max_connections": 20},  # Duplicate
                    ]
                }
            )
        )

        with pytest.raises(
            ValueError, match=r"Duplicate checkpointer 'type': 'sqlite'"
        ):
            await BatchCheckpointerConfig.from_yaml(file_path=file_path)

    @pytest.mark.asyncio
    async def test_missing_key_raises_error(self, temp_dir):
        """Test that missing required key raises an error."""
        file_path = temp_dir / "config.llms.yml"
        file_path.write_text(
            yaml.dump(
                {
                    "llms": [
                        {
                            "provider": "anthropic",  # Missing 'alias' key
                            "model": "claude-3-5-sonnet-20241022",
                            "max_tokens": 4096,
                            "temperature": 0.7,
                        }
                    ]
                }
            )
        )

        with pytest.raises(ValueError, match=r"missing required key 'alias'"):
            await BatchLLMConfig.from_yaml(file_path=file_path)

    @pytest.mark.asyncio
    async def test_unique_items_load_successfully(
        self, temp_dir, mock_llm_config, mock_checkpointer_config
    ):
        """Test that configs with unique items from file and directory load successfully."""
        file_path = temp_dir / "config.agents.yml"
        file_path.write_text(
            yaml.dump(
                {
                    "agents": [
                        {
                            "name": "agent1",
                            "default": True,
                            "llm": "test-model",
                            "checkpointer": "memory",
                        }
                    ]
                }
            )
        )

        dir_path = temp_dir / "agents"
        dir_path.mkdir()
        (dir_path / "agent2.yml").write_text(
            yaml.dump({"name": "agent2", "llm": "test-model"})
        )

        batch_llm = BatchLLMConfig(llms=[mock_llm_config])
        batch_checkpointer = BatchCheckpointerConfig(
            checkpointers=[mock_checkpointer_config]
        )

        config = await BatchAgentConfig.from_yaml(
            file_path=file_path,
            dir_path=dir_path,
            batch_llm_config=batch_llm,
            batch_checkpointer_config=batch_checkpointer,
        )

        assert len(config.agents) == 2
        assert set(config.agent_names) == {"agent1", "agent2"}


class TestFilenameValidation:
    """Test filename validation for directory-based configs."""

    @pytest.mark.asyncio
    async def test_agent_filename_mismatch_raises_error(
        self, temp_dir, mock_llm_config
    ):
        """Test that agent filename must match agent name."""
        dir_path = temp_dir / "agents"
        dir_path.mkdir()
        (dir_path / "wrong-name.yml").write_text(
            yaml.dump({"name": "correct-name", "llm": "test-model", "default": True})
        )

        batch_llm = BatchLLMConfig(llms=[mock_llm_config])

        with pytest.raises(
            ValueError,
            match=r"Agent file 'wrong-name.yml' has name='correct-name' but filename is 'wrong-name'",
        ):
            await BatchAgentConfig.from_yaml(
                dir_path=dir_path, batch_llm_config=batch_llm
            )

    @pytest.mark.asyncio
    async def test_checkpointer_filename_mismatch_raises_error(self, temp_dir):
        """Test that checkpointer filename must match checkpointer type."""
        dir_path = temp_dir / "checkpointers"
        dir_path.mkdir()
        (dir_path / "wrong-type.yml").write_text(
            yaml.dump({"type": "sqlite", "max_connections": 10})
        )

        with pytest.raises(
            ValueError,
            match=r"Checkpointer file 'wrong-type.yml' has type='sqlite' but filename is 'wrong-type'",
        ):
            await BatchCheckpointerConfig.from_yaml(dir_path=dir_path)

    @pytest.mark.asyncio
    async def test_matching_filenames_load_successfully(
        self, temp_dir, mock_llm_config
    ):
        """Test that correctly named files load successfully."""
        # Agents
        agents_dir = temp_dir / "agents"
        agents_dir.mkdir()
        (agents_dir / "my-agent.yml").write_text(
            yaml.dump({"name": "my-agent", "llm": "test-model", "default": True})
        )

        # Checkpointers
        checkpointers_dir = temp_dir / "checkpointers"
        checkpointers_dir.mkdir()
        (checkpointers_dir / "memory.yml").write_text(
            yaml.dump({"type": "memory", "max_connections": 10})
        )

        # All should load successfully
        agents_config = await BatchAgentConfig.from_yaml(
            dir_path=agents_dir, batch_llm_config=BatchLLMConfig(llms=[mock_llm_config])
        )
        checkpointers_config = await BatchCheckpointerConfig.from_yaml(
            dir_path=checkpointers_dir
        )

        assert len(agents_config.agents) == 1
        assert len(checkpointers_config.checkpointers) == 1



"""Tests for MessageContentBuilder."""

from src.cli.builders.message import MessageContentBuilder
from src.cli.resolvers import RefType


class TestExtractReferences:
    """Tests for extract_references method."""

    def test_extract_single_reference(self, temp_dir):
        """Test extracting a single @:image: reference."""
        builder = MessageContentBuilder(temp_dir)
        content = "Look at this @:image:photo.png image"

        refs = builder.extract_references(content)

        assert RefType.IMAGE in refs
        assert refs[RefType.IMAGE] == ["photo.png"]

    def test_extract_multiple_references(self, temp_dir):
        """Test extracting multiple @:image: references."""
        builder = MessageContentBuilder(temp_dir)
        content = "Compare @:image:photo1.png and @:image:photo2.jpg"

        refs = builder.extract_references(content)

        assert RefType.IMAGE in refs
        assert refs[RefType.IMAGE] == ["photo1.png", "photo2.jpg"]

    def test_extract_with_absolute_path(self, temp_dir):
        """Test extracting reference with absolute path."""
        builder = MessageContentBuilder(temp_dir)
        content = "Check @:image:/Users/test/photo.png"

        refs = builder.extract_references(content)

        assert RefType.IMAGE in refs
        assert refs[RefType.IMAGE] == ["/Users/test/photo.png"]

    def test_no_references(self, temp_dir):
        """Test when there are no image references."""
        builder = MessageContentBuilder(temp_dir)
        content = "Just some text without images"

        refs = builder.extract_references(content)

        assert refs == {}

    def test_reference_at_start(self, temp_dir):
        """Test reference at start of content."""
        builder = MessageContentBuilder(temp_dir)
        content = "@:image:photo.png is interesting"

        refs = builder.extract_references(content)

        assert RefType.IMAGE in refs
        assert refs[RefType.IMAGE] == ["photo.png"]

    def test_reference_at_end(self, temp_dir):
        """Test reference at end of content."""
        builder = MessageContentBuilder(temp_dir)
        content = "Look at @:image:photo.png"

        refs = builder.extract_references(content)

        assert RefType.IMAGE in refs
        assert refs[RefType.IMAGE] == ["photo.png"]

    def test_extract_absolute_path(self, temp_dir, create_test_image):
        """Test extracting standalone absolute path to image."""
        builder = MessageContentBuilder(temp_dir)
        image_path = create_test_image("photo")
        content = f"Look at this image: {image_path}"

        refs = builder.extract_references(content)

        assert RefType.IMAGE in refs
        assert str(image_path) in refs[RefType.IMAGE]

    def test_skip_image_references(self, temp_dir, create_test_image):
        """Test that @:image: references are counted separately."""
        builder = MessageContentBuilder(temp_dir)
        image_path = create_test_image("photo")
        content = f"@:image:{image_path} and {image_path}"

        refs = builder.extract_references(content)

        assert RefType.IMAGE in refs
        assert len(refs[RefType.IMAGE]) == 2

    def test_multiple_paths(self, temp_dir, create_test_image):
        """Test extracting multiple standalone paths."""
        builder = MessageContentBuilder(temp_dir)
        img1 = create_test_image("photo1")
        img2 = create_test_image("photo2", ".jpg")
        content = f"Compare {img1} with {img2}"

        refs = builder.extract_references(content)

        assert RefType.IMAGE in refs
        assert str(img1) in refs[RefType.IMAGE]
        assert str(img2) in refs[RefType.IMAGE]

    def test_no_image_paths(self, temp_dir):
        """Test when there are no image paths."""
        builder = MessageContentBuilder(temp_dir)
        content = "Just text without image paths"

        refs = builder.extract_references(content)

        assert refs == {}

    def test_relative_path_rejected(self, temp_dir):
        """Test that relative paths are not extracted."""
        builder = MessageContentBuilder(temp_dir)
        content = "Check out photo.png"

        refs = builder.extract_references(content)

        assert refs == {}


class TestBuild:
    """Tests for build method."""

    def test_build_with_text_and_image(self, temp_dir, create_test_image):
        """Test build with text and one image."""
        builder = MessageContentBuilder(temp_dir)
        image_path = create_test_image("photo")

        content, ref_mapping = builder.build(
            f"Describe this image @:image:{image_path}"
        )

        assert isinstance(content, list)
        assert len(content) == 2
        assert isinstance(content[0], dict) and content[0]["type"] == "text"
        assert (
            isinstance(content[0], dict) and "Describe this image" in content[0]["text"]
        )
        assert isinstance(content[1], dict) and content[1]["type"] == "image"
        assert isinstance(content[1], dict) and content[1]["source_type"] == "base64"

    def test_build_with_multiple_images(self, temp_dir, create_test_image):
        """Test build with multiple images."""
        builder = MessageContentBuilder(temp_dir)
        img1 = create_test_image("photo1")
        img2 = create_test_image("photo2")

        content, ref_mapping = builder.build(
            f"Compare @:image:{img1} and @:image:{img2}"
        )

        assert isinstance(content, list)
        assert len(content) == 3
        assert isinstance(content[0], dict) and content[0]["type"] == "text"
        assert isinstance(content[1], dict) and content[1]["type"] == "image"
        assert isinstance(content[2], dict) and content[2]["type"] == "image"

    def test_build_only_images(self, temp_dir, create_test_image):
        """Test build with only images."""
        builder = MessageContentBuilder(temp_dir)
        image_path = create_test_image("photo")

        content, ref_mapping = builder.build(f"@:image:{image_path}")

        assert isinstance(content, list)
        assert len(content) == 1
        assert isinstance(content[0], dict) and content[0]["type"] == "image"

    def test_build_integration(self, temp_dir, create_test_image):
        """Test build handles whitespace correctly."""
        builder = MessageContentBuilder(temp_dir)
        image_path = create_test_image("photo")

        content, ref_mapping = builder.build(
            f"  \n  Text with whitespace  \n  @:image:{image_path}"
        )

        assert isinstance(content, list)
        text_blocks = [
            b for b in content if isinstance(b, dict) and b.get("type") == "text"
        ]
        assert len(text_blocks) == 1
        assert (
            isinstance(text_blocks[0], dict)
            and "Text with whitespace" in text_blocks[0]["text"]
        )

    def test_build_text_only(self, temp_dir):
        """Test build with text only returns string."""
        builder = MessageContentBuilder(temp_dir)
        content, ref_mapping = builder.build("Just text")

        assert isinstance(content, str)
        assert content == "Just text"
        assert ref_mapping == {}

    def test_build_nonexistent_image_raises_error(self, temp_dir):
        """Test build raises error for non-existent image."""
        import pytest

        builder = MessageContentBuilder(temp_dir)

        with pytest.raises(ValueError, match="Image not found"):
            builder.build("@:image:nonexistent.png")

    def test_build_unsupported_format_raises_error(self, temp_dir):
        """Test build raises error for unsupported image format."""
        import pytest

        # Create a file with unsupported extension
        test_file = temp_dir / "test.txt"
        test_file.write_text("not an image")

        builder = MessageContentBuilder(temp_dir)

        with pytest.raises(ValueError, match="Unsupported format"):
            builder.build(f"@:image:{test_file}")



"""Tests for chat command handler."""

from pathlib import Path
from unittest.mock import patch

import pytest

from src.cli.bootstrap.chat import handle_chat_command


class TestHandleChatCommand:
    """Tests for handle_chat_command function."""

    @pytest.mark.asyncio
    async def test_handle_chat_command_creates_context(
        self, mock_app_args, patch_chat_dependencies
    ):
        """Test that handle_chat_command creates a context."""
        result = await handle_chat_command(mock_app_args)
        assert result == 0

    @pytest.mark.asyncio
    async def test_handle_chat_command_starts_session(
        self, mock_app_args, patch_chat_dependencies
    ):
        """Test that handle_chat_command starts a CLI session."""
        result = await handle_chat_command(mock_app_args)

        patch_chat_dependencies["session"].start.assert_called_once()
        assert result == 0

    @pytest.mark.asyncio
    async def test_handle_chat_command_with_resume_flag(
        self, mock_app_args, patch_chat_dependencies, mock_context
    ):
        """Test that handle_chat_command handles resume flag in interactive mode."""
        mock_app_args.resume = True

        result = await handle_chat_command(mock_app_args)

        patch_chat_dependencies[
            "session"
        ].command_dispatcher.resume_handler.handle.assert_called_once_with(
            mock_context.thread_id
        )
        assert result == 0

    @pytest.mark.asyncio
    async def test_handle_chat_command_enables_timer(
        self, mock_app_args, patch_chat_dependencies
    ):
        """Test that handle_chat_command enables timer when requested."""
        mock_app_args.timer = True

        result = await handle_chat_command(mock_app_args)

        patch_chat_dependencies["enable_timer"].assert_called_once()
        assert result == 0

    @pytest.mark.asyncio
    async def test_handle_chat_command_handles_reload_loop(
        self, mock_app_args, patch_chat_dependencies
    ):
        """Test that handle_chat_command handles reload loop."""
        call_count = 0
        session = patch_chat_dependencies["session"]

        def side_effect(*_args, **_kwargs):
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                session.needs_reload = True
            else:
                session.needs_reload = False

        session.start.side_effect = side_effect

        result = await handle_chat_command(mock_app_args)

        assert session.start.call_count == 2
        assert result == 0

    @pytest.mark.asyncio
    async def test_handle_chat_command_shows_welcome_on_first_start(
        self, mock_app_args, patch_chat_dependencies
    ):
        """Test that handle_chat_command shows welcome message on first start."""
        await handle_chat_command(mock_app_args)

        patch_chat_dependencies["session"].start.assert_called_once_with(
            show_welcome=True
        )

    @pytest.mark.asyncio
    async def test_handle_chat_command_hides_welcome_on_resume(
        self, mock_app_args, patch_chat_dependencies
    ):
        """Test that handle_chat_command hides welcome when resuming."""
        mock_app_args.resume = True

        await handle_chat_command(mock_app_args)

        patch_chat_dependencies["session"].start.assert_called_once_with(
            show_welcome=False
        )

    @pytest.mark.asyncio
    async def test_handle_chat_command_handles_keyboard_interrupt(
        self, mock_app_args, patch_chat_dependencies
    ):
        """Test that handle_chat_command handles KeyboardInterrupt gracefully."""
        patch_chat_dependencies["session"].start.side_effect = KeyboardInterrupt()

        result = await handle_chat_command(mock_app_args)

        assert result == 0

    @pytest.mark.asyncio
    async def test_handle_chat_command_handles_exception(
        self, mock_app_args, patch_chat_dependencies
    ):
        """Test that handle_chat_command handles exceptions and returns error code."""
        patch_chat_dependencies["session"].start.side_effect = Exception("Test error")

        result = await handle_chat_command(mock_app_args)

        assert result == 1

    @pytest.mark.asyncio
    async def test_handle_chat_command_uses_path_object(
        self, mock_app_args, patch_chat_dependencies
    ):
        """Test that handle_chat_command converts working_dir to Path."""
        await handle_chat_command(mock_app_args)

        call_args = patch_chat_dependencies["context_create"].call_args
        assert isinstance(call_args[1]["working_dir"], Path)

    @pytest.mark.asyncio
    async def test_handle_chat_command_one_shot_mode(
        self, mock_app_args, patch_chat_dependencies
    ):
        """Test that handle_chat_command sends a single message in one-shot mode."""
        mock_app_args.message = "test message"

        result = await handle_chat_command(mock_app_args)

        patch_chat_dependencies["session"].send.assert_called_once_with("test message")
        patch_chat_dependencies["session"].start.assert_not_called()
        assert result == 0

    @pytest.mark.asyncio
    async def test_handle_chat_command_one_shot_mode_with_resume(
        self, mock_app_args, patch_chat_dependencies, mock_context
    ):
        """Test that handle_chat_command handles resume in one-shot mode without rendering history."""
        mock_app_args.message = "test message"
        mock_app_args.resume = True

        result = await handle_chat_command(mock_app_args)

        patch_chat_dependencies[
            "session"
        ].command_dispatcher.resume_handler.handle.assert_called_once_with(
            mock_context.thread_id, render_history=False
        )
        patch_chat_dependencies["session"].send.assert_called_once_with("test message")
        patch_chat_dependencies["session"].start.assert_not_called()
        assert result == 0


@pytest.fixture
def patch_chat_dependencies(mock_context, mock_session):
    """Patch Context.create and Session for chat tests."""
    with (
        patch(
            "src.cli.bootstrap.chat.Context.create", return_value=mock_context
        ) as mock_create,
        patch(
            "src.cli.bootstrap.chat.Session", return_value=mock_session
        ) as mock_session_cls,
        patch("src.cli.bootstrap.chat.enable_timer") as mock_enable_timer,
    ):
        yield {
            "context_create": mock_create,
            "session_cls": mock_session_cls,
            "session": mock_session,
            "enable_timer": mock_enable_timer,
        }



from pathlib import Path

from pydantic import BaseModel

from src.core.config import ApprovalMode


class AgentContext(BaseModel):
    """Runtime context for agent execution."""

    approval_mode: ApprovalMode
    working_dir: Path
    input_cost_per_mtok: float | None = None
    output_cost_per_mtok: float | None = None
    tool_output_max_tokens: int | None = None



"""Slash command parser and handlers."""

import shlex
import uuid
from collections.abc import Callable

from src.cli.bootstrap.initializer import initializer
from src.cli.handlers import (
    AgentHandler,
    CompressionHandler,
    GraphHandler,
    MCPHandler,
    MemoryHandler,
    ModelHandler,
    ReplayHandler,
    ResumeHandler,
    ToolsHandler,
)
from src.cli.theme import console


class CommandDispatcher:
    """Dispatch slash commands."""

    def __init__(self, session):
        """Initialize with reference to CLI session."""
        self.session = session
        self.commands = self._register_commands()
        self.resume_handler = ResumeHandler(session)
        self.agent_handler = AgentHandler(session)
        self.model_handler = ModelHandler(session)
        self.mcp_handler = MCPHandler(session)
        self.memory_handler = MemoryHandler(session)
        self.tools_handler = ToolsHandler(session)
        self.replay_handler = ReplayHandler(session)
        self.compression_handler = CompressionHandler(session)
        self.graph_handler = GraphHandler(session)

    def _register_commands(self) -> dict[str, Callable]:
        """Register all available commands."""
        return {
            "/help": self.cmd_help,
            "/hotkeys": self.cmd_hotkeys,
            "/agents": self.cmd_agents,
            "/model": self.cmd_model,
            "/tools": self.cmd_tools,
            "/mcp": self.cmd_mcp,
            "/memory": self.cmd_memory,
            "/graph": self.cmd_graph,
            "/clear": self.cmd_clear,
            "/exit": self.cmd_exit,
            "/resume": self.cmd_resume,
            "/replay": self.cmd_replay,
            "/compress": self.cmd_compress,
        }

    async def dispatch(self, command_line: str) -> None:
        """Dispatch a slash command."""
        if not command_line.startswith("/"):
            console.print_error("Commands must start with '/'")
            console.print("")
            return

        try:
            parts = shlex.split(command_line)
            if not parts:
                console.print_error("Empty command")
                console.print("")
                return

            command = parts[0].lower()
            args = parts[1:] if len(parts) > 1 else []

            if command in self.commands:
                await self.commands[command](args)
            else:
                console.print_error(f"Unknown command: {command}")
                console.print("")
                await self.cmd_help([])

        except Exception as e:
            console.print_error(f"Command error: {e}")
            console.print("")

    async def cmd_help(self, args: list[str]) -> None:
        """Show help information."""
        self.session.renderer.render_help(self.commands)

    async def cmd_hotkeys(self, args: list[str]) -> None:
        """Show keyboard shortcuts."""
        self.session.renderer.render_hotkeys(self.session.prompt.hotkeys)

    async def cmd_agents(self, args: list[str]) -> None:
        """Handle agents command with interactive selector."""
        await self.agent_handler.handle()

    async def cmd_model(self, args: list[str]) -> None:
        """Handle model command with interactive selector."""
        await self.model_handler.handle()

    async def cmd_tools(self, args: list[str]) -> None:
        """Handle tools command with interactive selector."""
        await self.tools_handler.handle(initializer.cached_tools)

    async def cmd_mcp(self, args: list[str]) -> None:
        """Handle MCP management command."""
        await self.mcp_handler.handle()

    async def cmd_memory(self, args: list[str]) -> None:
        """Open memory file for editing user preferences and context."""
        await self.memory_handler.handle()

    async def cmd_clear(self, args: list[str]) -> None:
        """Clear the screen and start a new thread."""
        new_thread_id = str(uuid.uuid4())
        self.session.update_context(
            thread_id=new_thread_id,
            current_input_tokens=None,
            current_output_tokens=None,
            total_cost=None,
        )
        console.clear()

    async def cmd_exit(self, args: list[str]) -> None:
        """Exit the application."""
        self.session.running = False

    async def cmd_resume(self, args: list[str]) -> None:
        """Resume conversation thread with interactive selector."""
        await self.resume_handler.handle()

    async def cmd_replay(self, args: list[str]) -> None:
        """Replay conversation from a previous human message."""
        await self.replay_handler.handle()

    async def cmd_compress(self, args: list[str]) -> None:
        """Compress conversation history to a new thread."""
        await self.compression_handler.handle()

    async def cmd_graph(self, args: list[str]) -> None:
        """Render and display the current agent graph (use --browser to open in browser)."""
        # Validate and parse arguments
        if invalid_args := [arg for arg in args if arg != "--browser"]:
            console.print_error(f"Invalid argument(s): {', '.join(invalid_args)}")
            console.print("[muted]Usage: /graph [--browser][/muted]", markup=True)
            console.print("")
            return

        await self.graph_handler.handle(open_browser="--browser" in args)



"""Base checkpointer with extended methods for langrepl."""

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from langchain_core.messages import BaseMessage
from langgraph.checkpoint.base import BaseCheckpointSaver as _BaseCheckpointSaver
from langgraph.checkpoint.base import CheckpointTuple


@dataclass
class HumanMessageEntry:
    """Human message with replay metadata."""

    text: str
    reference_mapping: dict[str, Any]
    messages_before_count: int
    checkpoint_id: str | None
    input_tokens: int | None = None
    output_tokens: int | None = None
    total_cost: float | None = None


class BaseCheckpointer(_BaseCheckpointSaver):
    """Base checkpointer with additional query methods."""

    async def get_threads(self) -> set[str]:
        """Get all thread IDs."""
        raise NotImplementedError

    async def get_history(self, latest: CheckpointTuple) -> list[CheckpointTuple]:
        """Get checkpoint history in chronological order (oldest first)."""
        raise NotImplementedError

    async def delete_after(self, thread_id: str, checkpoint_id: str | None) -> int:
        """Delete checkpoints after checkpoint_id. Returns count deleted."""
        raise NotImplementedError

    async def get_human_messages(
        self,
        thread_id: str,
        latest: CheckpointTuple,
        on_indexing: Callable[[], None] | None = None,
    ) -> tuple[list[HumanMessageEntry], list[BaseMessage]]:
        """Get human messages with replay metadata.

        Returns:
            Tuple of (human_messages, all_messages)
        """
        raise NotImplementedError



"""Slash command completer."""

from collections.abc import AsyncGenerator, Iterable

from prompt_toolkit.completion import CompleteEvent, Completion, WordCompleter
from prompt_toolkit.document import Document


class SlashCommandCompleter(WordCompleter):
    """Auto-completer for slash commands."""

    def __init__(self, commands: list[str]):
        super().__init__(commands, ignore_case=True, sentence=True)

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        """Get completions for slash commands (sync)."""
        yield from super().get_completions(document, complete_event)

    async def get_completions_async(
        self, document: Document, complete_event: CompleteEvent
    ) -> AsyncGenerator[Completion]:
        """Get completions for slash commands (async)."""
        for completion in self.get_completions(document, complete_event):
            yield completion



"""Tests for FileResolver."""

from pathlib import Path
from shlex import quote
from unittest.mock import patch

import pytest

from src.cli.resolvers.file import FileResolver


class TestFileResolverGetFiles:
    """Tests for FileResolver._get_files method."""

    @pytest.mark.asyncio
    @patch("src.cli.resolvers.file.execute_bash_command")
    async def test_get_files_with_git(self, mock_exec, temp_dir):
        """Test getting files using git."""
        mock_exec.return_value = (0, "file1.py\nfile2.py\nfile3.py\n", "")

        files = await FileResolver._get_files(temp_dir)

        assert files == ["file1.py", "file2.py", "file3.py"]
        assert "git ls-files" in mock_exec.call_args[0][0][2]

    @pytest.mark.asyncio
    @patch("src.cli.resolvers.file.execute_bash_command")
    async def test_get_files_git_fails_fallback_to_fd(self, mock_exec, temp_dir):
        """Test fallback to fd when git fails."""
        mock_exec.side_effect = [
            (1, "", "not a git repository"),
            (0, "file1.py\nfile2.py\n", ""),
        ]

        files = await FileResolver._get_files(temp_dir)

        assert files == ["file1.py", "file2.py"]
        assert mock_exec.call_count == 2
        assert "fd --type f" in mock_exec.call_args_list[1][0][0][2]

    @pytest.mark.asyncio
    @patch("src.cli.resolvers.file.execute_bash_command")
    async def test_get_files_both_commands_fail(self, mock_exec, temp_dir):
        """Test when both git and fd fail."""
        mock_exec.side_effect = [
            (1, "", "not a git repository"),
            (1, "", "fd not found"),
        ]

        files = await FileResolver._get_files(temp_dir)

        assert files == []

    @pytest.mark.asyncio
    @patch("src.cli.resolvers.file.execute_bash_command")
    async def test_get_files_filters_empty_lines(self, mock_exec, temp_dir):
        """Test that empty lines are filtered out."""
        mock_exec.return_value = (0, "file1.py\n\nfile2.py\n\n", "")

        files = await FileResolver._get_files(temp_dir)

        assert files == ["file1.py", "file2.py"]

    @pytest.mark.asyncio
    @patch("src.cli.resolvers.file.execute_bash_command")
    async def test_get_files_pattern_escaping(self, mock_exec, temp_dir):
        """Test that pattern is properly quoted for shell safety."""
        mock_exec.return_value = (0, "test.py\n", "")

        await FileResolver._get_files(temp_dir, pattern="$(malicious)")

        call_args = mock_exec.call_args[0][0][2]
        assert quote("$(malicious)") in call_args


class TestFileResolverGetDirectories:
    """Tests for FileResolver._get_directories method."""

    @pytest.mark.asyncio
    @patch("src.cli.resolvers.file.execute_bash_command")
    async def test_get_directories_filters_dot(self, mock_exec, temp_dir):
        """Test that current directory (.) is filtered out."""
        mock_exec.return_value = (0, ".\nsrc\ntests\n", "")

        dirs = await FileResolver._get_directories(temp_dir)

        assert "." not in dirs
        assert dirs == ["src", "tests"]

    @pytest.mark.asyncio
    @patch("src.cli.resolvers.file.execute_bash_command")
    async def test_get_directories_git_fails_fallback_to_fd(self, mock_exec, temp_dir):
        """Test fallback to fd when git fails."""
        mock_exec.side_effect = [
            (1, "", "not a git repository"),
            (0, "src\ntests\n", ""),
        ]

        dirs = await FileResolver._get_directories(temp_dir)

        assert dirs == ["src", "tests"]
        assert mock_exec.call_count == 2
        assert "fd --type d" in mock_exec.call_args_list[1][0][0][2]


class TestFileResolverResolve:
    """Tests for FileResolver.resolve method."""

    def test_resolve_relative_path(self, temp_dir):
        """Test resolving relative path."""
        resolver = FileResolver()
        ctx = {"working_dir": str(temp_dir)}

        result = resolver.resolve("test.py", ctx)

        expected = str((temp_dir / "test.py").resolve())
        assert result == expected

    def test_resolve_nested_relative_path(self, temp_dir):
        """Test resolving nested relative path."""
        resolver = FileResolver()
        ctx = {"working_dir": str(temp_dir)}

        result = resolver.resolve("src/test.py", ctx)

        expected = str((temp_dir / "src" / "test.py").resolve())
        assert result == expected

    def test_resolve_absolute_path(self, temp_dir):
        """Test resolving absolute path."""
        resolver = FileResolver()
        ctx = {"working_dir": str(temp_dir)}
        absolute_path = "/tmp/test.py"

        result = resolver.resolve(absolute_path, ctx)

        assert result == str(Path(absolute_path).resolve())

    def test_resolve_root_slash(self, temp_dir):
        """Test resolving root slash returns working directory."""
        resolver = FileResolver()
        ctx = {"working_dir": str(temp_dir)}

        result = resolver.resolve("/", ctx)

        assert result == str(temp_dir.resolve())

    def test_resolve_parent_directory_reference(self, temp_dir):
        """Test resolving parent directory reference."""
        resolver = FileResolver()
        ctx = {"working_dir": str(temp_dir)}

        result = resolver.resolve("../test.py", ctx)

        expected = str((temp_dir / ".." / "test.py").resolve())
        assert result == expected

    def test_resolve_with_invalid_working_dir_returns_original(self):
        """Test resolve returns original ref when path resolution fails."""
        resolver = FileResolver()
        ctx = {"working_dir": ""}

        with patch("src.cli.resolvers.file.resolve_path") as mock_resolve:
            mock_resolve.side_effect = Exception("Invalid path")
            result = resolver.resolve("test.py", ctx)

        assert result == "test.py"


class TestFileResolverComplete:
    """Tests for FileResolver.complete method."""

    @pytest.mark.asyncio
    @patch("src.cli.resolvers.file.execute_bash_command")
    async def test_complete_returns_files_and_directories(self, mock_exec, temp_dir):
        """Test complete returns both files and directories."""
        resolver = FileResolver()
        ctx = {"working_dir": str(temp_dir), "start_position": 0}

        mock_exec.side_effect = [
            (0, "file1.py\nfile2.py\n", ""),
            (0, "src\ntests\n", ""),
        ]

        completions = await resolver.complete("", ctx, limit=10)

        assert len(completions) == 4
        completion_texts = [c.text for c in completions]
        assert "@:file:file1.py" in completion_texts
        assert "@:file:file2.py" in completion_texts
        assert "@:file:src" in completion_texts
        assert "@:file:tests" in completion_texts

    @pytest.mark.asyncio
    @patch("src.cli.resolvers.file.execute_bash_command")
    async def test_complete_directories_have_trailing_slash_in_display(
        self, mock_exec, temp_dir
    ):
        """Test directories have trailing slash in display text."""
        resolver = FileResolver()
        ctx = {"working_dir": str(temp_dir), "start_position": 0}

        mock_exec.side_effect = [
            (0, "file.py\n", ""),
            (0, "src\n", ""),
        ]

        completions = await resolver.complete("", ctx, limit=10)

        dir_completion = next(c for c in completions if "src" in c.text)
        file_completion = next(c for c in completions if "file.py" in c.text)

        assert "@:file:src/" in str(dir_completion.display)
        assert dir_completion.text == "@:file:src"
        assert "@:file:file.py" in str(file_completion.display)
        assert file_completion.text == "@:file:file.py"

    @pytest.mark.asyncio
    @patch("src.cli.resolvers.file.execute_bash_command")
    async def test_complete_sorts_directories_before_files_in_same_parent(
        self, mock_exec, temp_dir
    ):
        """Test directories appear before files in same parent directory."""
        resolver = FileResolver()
        ctx = {"working_dir": str(temp_dir), "start_position": 0}

        mock_exec.side_effect = [
            (0, "src/file.py\nroot.py\n", ""),
            (0, "src\n", ""),
        ]

        completions = await resolver.complete("", ctx, limit=10)

        completion_texts = [c.text for c in completions]
        src_dir_index = completion_texts.index("@:file:src")
        root_file_index = completion_texts.index("@:file:root.py")
        src_file_index = completion_texts.index("@:file:src/file.py")

        assert src_dir_index < root_file_index < src_file_index

    @pytest.mark.asyncio
    @patch("src.cli.resolvers.file.execute_bash_command")
    async def test_complete_handles_exception_gracefully(self, mock_exec, temp_dir):
        """Test complete returns empty list when bash command raises exception."""
        resolver = FileResolver()
        ctx = {"working_dir": str(temp_dir), "start_position": 0}

        mock_exec.side_effect = Exception("Command failed")

        completions = await resolver.complete("", ctx, limit=10)

        assert completions == []

    @pytest.mark.asyncio
    @patch("src.cli.resolvers.file.execute_bash_command")
    async def test_complete_with_empty_results(self, mock_exec, temp_dir):
        """Test complete with no files or directories found."""
        resolver = FileResolver()
        ctx = {"working_dir": str(temp_dir), "start_position": 0}

        mock_exec.side_effect = [
            (0, "", ""),
            (0, "", ""),
        ]

        completions = await resolver.complete("nonexistent", ctx, limit=10)

        assert completions == []
